/*
 * Device driver for the SYMBIOS/LSILOGIC 53C8XX and 53C1010 family 
 * of PCI-SCSI IO processors.
 *
 * Copyright (C) 1999-2001  Gerard Roudier <groudier@free.fr>
 *
 * This driver is derived from the Linux sym53c8xx driver.
 * Copyright (C) 1998-2000  Gerard Roudier
 *
 * The sym53c8xx driver is derived from the ncr53c8xx driver that had been 
 * a port of the FreeBSD ncr driver to Linux-1.2.13.
 *
 * The original ncr driver has been written for 386bsd and FreeBSD by
 *         Wolfgang Stanglmeier        <wolf@cologne.de>
 *         Stefan Esser                <se@mi.Uni-Koeln.de>
 * Copyright (C) 1994  Wolfgang Stanglmeier
 *
 * Other major contributions:
 *
 * NVRAM detection and reading.
 * Copyright (C) 1997 Richard Waltham <dormouse@farsrobt.demon.co.uk>
 *
 *-----------------------------------------------------------------------------
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifdef __FreeBSD__
#include <dev/sym/sym_glue.h>
#else
#include "sym_glue.h"
#endif

/*
 *  Some mappings to PCI configuration access functions.
 *
 *  This allows to get rid of some #ifdef in the driver code and 
 *  makes the code slightly more clear.
 */
#ifdef FreeBSD_Bus_Io_Abstraction
#define sym_pci_read_config   pci_read_config
#define sym_pci_write_config  pci_write_config
#define sym_pci_get_vendor    pci_get_vendor
#define sym_pci_get_device    pci_get_device
#define sym_pci_get_revid     pci_get_revid
#else
#define sym_pci_read_config   pci_cfgread
#define sym_pci_write_config  pci_cfgwrite
#define sym_pci_get_vendor(d) pci_cfgread((d), PCIR_VENDOR, 2)
#define sym_pci_get_device(d) pci_cfgread((d), PCIR_DEVICE, 2)
#define sym_pci_get_revid(d)  pci_cfgread((d), PCIR_REVID,  1)
#define sym_pci_get_bus(d)   ((d)->bus)
#define sym_pci_get_slot(d)  ((d)->slot)
#endif

#define sym_pci_read_config_u8(d, r)      sym_pci_read_config((d), (r), 1)
#define sym_pci_read_config_u16(d, r)     sym_pci_read_config((d), (r), 2)
#define sym_pci_read_config_u32(d, r)     sym_pci_read_config((d), (r), 4)
#define sym_pci_write_config_u8(d, r, v)  sym_pci_write_config((d), (r), (v), 1)
#define sym_pci_write_config_u16(d, r, v) sym_pci_write_config((d), (r), (v), 2)
#define sym_pci_write_config_u32(d, r, v) sym_pci_write_config((d), (r), (v), 4)

/*
 *  BARs offset.
 */
#define SYM_PCI_IO	PCIR_MAPS
#define SYM_PCI_MMIO	(PCIR_MAPS + 4)
#define SYM_PCI_RAM	(PCIR_MAPS + 8)
#define SYM_PCI_RAM64	(PCIR_MAPS + 12)

/*
 *  Map/unmap chip bus resources.
 *
 *  Interfaces seem so uselessly complex to me, that I donnot 
 *  want to use them more than once. Thus these mappings. :)
 */
#ifdef FreeBSD_Bus_Io_Abstraction

static u32 
sym_map_bus_mem(sym_device_t dev, int regs_id, struct sym_bus_mem *mmio)
{
	mmio->reg = regs_id;
	mmio->res = bus_alloc_resource(dev, SYS_RES_MEMORY, &mmio->reg,
				       0, ~0, 1, RF_ACTIVE);
	if (!mmio->res)
		return 0;
	mmio->bsh = rman_get_bushandle(mmio->res);
	mmio->tag = rman_get_bustag(mmio->res);
	return (u32)rman_get_start(mmio->res);
}
static void __inline 
sym_unmap_bus_mem(sym_device_t dev, struct sym_bus_mem *mmio)
{
	if (mmio->res) {
		bus_release_resource(dev, SYS_RES_MEMORY, mmio->reg, mmio->res);
		mmio->res = 0;
	}
}
#ifdef	SYM_CONF_IOMAPPED
static u32 __inline
sym_map_bus_io(sym_device_t dev, int regs_id, struct sym_bus_io *io)
{
	io->reg = regs_id;
	io->res = bus_alloc_resource(dev, SYS_RES_IOPORT, &io->reg,
				     0, ~0, 1, RF_ACTIVE);
	if (!io->res)
		return 0;
	io->bsh  = rman_get_bushandle(io->res);
	io->tag  = rman_get_bustag(io->res);
	return (u32)rman_get_start(io->res);
}
static void __inline 
sym_unmap_bus_io(sym_device_t dev, struct sym_bus_io *io)
{
	if (io->res) {
		bus_release_resource(dev, SYS_RES_IOPORT, io->reg, io->res);
		io->res = 0;
	}
}
#endif	/* SYM_CONF_IOMAPPED */

#else	/* !def FreeBSD_Bus_Io_Abstraction */

static u32 __inline 
sym_map_bus_mem(sym_device_t dev, int regs_id, struct sym_bus_mem *mmio)
{
	vm_offset_t pa;
	if (!pci_map_mem(dev, regs_id, &mmio->va, &pa))
		pa = 0;
	return (u32)pa;
}
#define sym_unmap_bus_mem(dev, mmio)	do { ; } while (0)
#ifdef	SYM_CONF_IOMAPPED
static u32 __inline
sym_map_bus_io(sym_device_t dev, int regs_id, struct sym_bus_io *io)
{
	if (!pci_map_port (dev, SYM_PCI_IO, &io->port))
		io->port = 0;
	return (u32)io->port;
}
#define sym_unmap_bus_io(dev, io)	do { ; } while (0)
#endif	/* SYM_CONF_IOMAPPED */

#endif /* FreeBSD_Bus_Io_Abstraction */

/*
 *  Needed function prototypes.
 */
static void sym_timeout (void *arg);
static void sym_action (struct cam_sim *sim, union ccb *ccb);
static void sym_action1 (struct cam_sim *sim, union ccb *ccb);

#ifdef	FreeBSD_Bus_Dma_Abstraction
static int sym_fast_scatter_sg_physical(hcb_p np, ccb_p cp, 
					bus_dma_segment_t *psegs, int nsegs);
#else
static int  sym_scatter_virtual (hcb_p np, ccb_p cp, vm_offset_t vaddr,
				 vm_size_t len);
static int  sym_scatter_sg_virtual (hcb_p np, ccb_p cp, 
				    bus_dma_segment_t *psegs, int nsegs);
static int  sym_scatter_physical (hcb_p np, ccb_p cp, vm_offset_t paddr,
				  vm_size_t len);
#endif
static int sym_scatter_sg_physical (hcb_p np, ccb_p cp, 
				    bus_dma_segment_t *psegs, int nsegs);
static void sym_action2 (struct cam_sim *sim, union ccb *ccb);
static void sym_update_trans (hcb_p np, tcb_p tp, struct sym_trans *tip,
			      struct ccb_trans_settings *cts);
static void sym_update_dflags(hcb_p np, u_char *flags,
			      struct ccb_trans_settings *cts);

#ifdef FreeBSD_Bus_Io_Abstraction
static struct sym_pci_chip *sym_find_pci_chip (device_t dev);
static int  sym_pci_probe (device_t dev);
static int  sym_pci_attach (device_t dev);
#else
static struct sym_pci_chip *sym_find_pci_chip (pcici_t tag);
static const char *sym_pci_probe (pcici_t tag, pcidi_t type);
static void sym_pci_attach (pcici_t tag, int unit);
static int sym_pci_attach2 (pcici_t tag, int unit);
#endif

static void sym_pci_free (hcb_p np);
static int  sym_cam_attach (hcb_p np);
static void sym_cam_free (hcb_p np);

/*
 *  Insert a delay in micro-seconds and milli-seconds.
 */
void sym_udelay(int us) { DELAY(us); }
void sym_mdelay(int ms) { while (ms--) UDELAY(1000); }

/*
 *  Call-back used by the DMAable memory allocator.
 */
#ifdef FreeBSD_Bus_Dma_Abstraction
void sym_cb_dmamap_load(void *arg, bus_dma_segment_t *segs, int nseg, int err)
{
	bus_addr_t *baddr;
	baddr = (bus_addr_t *)arg;
	*baddr = segs->ds_addr;
}
#endif


/*
 *  Take into account this ccb in the freeze count.
 */	
void sym_freeze_cam_ccb(union ccb *ccb)
{
	if (!(ccb->ccb_h.flags & CAM_DEV_QFRZDIS)) {
		if (!(ccb->ccb_h.status & CAM_DEV_QFRZN)) {
			ccb->ccb_h.status |= CAM_DEV_QFRZN;
			xpt_freeze_devq(ccb->ccb_h.path, 1);
		}
	}
}

/*
 *  Enqueue a CAM CCB.
 */
static void sym_enqueue_cam_ccb(hcb_p np, union ccb *ccb)
{
	assert(!(ccb->ccb_h.status & CAM_SIM_QUEUED));
	ccb->ccb_h.status = CAM_REQ_INPROG;

	ccb->ccb_h.timeout_ch = timeout(sym_timeout, (caddr_t) ccb,
				       ccb->ccb_h.timeout*hz/1000);
	ccb->ccb_h.status |= CAM_SIM_QUEUED;
	ccb->ccb_h.sym_hcb_ptr = np;

	sym_insque_tail(sym_qptr(&ccb->ccb_h.sim_links), &np->s.cam_ccbq);
}

/*
 *  Complete a pending CAM CCB.
 */
void sym_xpt_done(hcb_p np, union ccb *ccb)
{
	if (ccb->ccb_h.status & CAM_SIM_QUEUED) {
		untimeout(sym_timeout, (caddr_t) ccb, ccb->ccb_h.timeout_ch);
		sym_remque(sym_qptr(&ccb->ccb_h.sim_links));
		ccb->ccb_h.status &= ~CAM_SIM_QUEUED;
		ccb->ccb_h.sym_hcb_ptr = 0;
	}
	if (ccb->ccb_h.flags & CAM_DEV_QFREEZE)
		sym_freeze_cam_ccb(ccb);
#ifdef DEBUG_2_0_X
	if (ccb->ccb_h.status != CAM_REQ_CMP)
		printf("XXXX STATUS=0x%x\n", ccb->ccb_h.status);
#endif
	xpt_done(ccb);
}

void sym_xpt_done2(hcb_p np, union ccb *ccb, int cam_status)
{
	sym_set_cam_status(ccb, cam_status);
	sym_xpt_done(np, ccb);
}

/*
 *  Print something that identifies the IO.
 */
void sym_print_addr (ccb_p cp)
{
	if (cp && cp->cam_ccb)
		xpt_print_path(cp->cam_ccb->ccb_h.path);
}

/*
 *  Tell the SCSI layer about a BUS RESET.
 */
void sym_xpt_async_bus_reset(hcb_p np)
{
	xpt_async(AC_BUS_RESET, np->s.path, NULL);
}

/*
 *  Tell the SCSI layer about a BUS DEVICE RESET message sent.
 */
void sym_xpt_async_sent_bdr(hcb_p np, int target)
{
	struct cam_path *path;
	int sts;

	sts = xpt_create_path(&path, NULL, cam_sim_path(np->s.sim), target,
	                      CAM_LUN_WILDCARD);
	if (sts != CAM_REQ_CMP)
		return;

	xpt_async(AC_SENT_BDR, path, NULL);
	xpt_free_path(path);
}

/*
 *  Tell the SCSI layer about the new transfer parameters.
 */
void __sym_xpt_async_transfer_neg(hcb_p np, int target, u_int spi_valid)
{
	struct ccb_trans_settings cts;
	struct cam_path *path;
	int sts;
	tcb_p tp = &np->target[target];

	sts = xpt_create_path(&path, NULL, cam_sim_path(np->s.sim), target,
	                      CAM_LUN_WILDCARD);
	if (sts != CAM_REQ_CMP)
		return;

	bzero(&cts, sizeof(cts));

#ifdef	FreeBSD_New_Tran_Settings
#define	cts__scsi (cts.proto_specific.scsi)
#define	cts__spi  (cts.xport_specific.spi)

	cts.type      = CTS_TYPE_CURRENT_SETTINGS;
	cts.protocol  = PROTO_SCSI;
	cts.transport = XPORT_SPI;
	cts.protocol_version  = tp->tinfo.curr.scsi_version;
	cts.transport_version = tp->tinfo.curr.spi_version;

	cts__spi.valid = spi_valid;
	if (spi_valid & CTS_SPI_VALID_SYNC_RATE)
		cts__spi.sync_period = tp->tinfo.curr.period;
	if (spi_valid & CTS_SPI_VALID_SYNC_OFFSET)
		cts__spi.sync_offset = tp->tinfo.curr.offset;
	if (spi_valid & CTS_SPI_VALID_BUS_WIDTH)
		cts__spi.bus_width   = tp->tinfo.curr.width;
	if (spi_valid & CTS_SPI_VALID_PPR_OPTIONS)
		cts__spi.ppr_options = tp->tinfo.curr.options;
#undef cts__spi
#undef cts__scsi
#else
	cts.valid = spi_valid;
	if (spi_valid & CCB_TRANS_SYNC_RATE_VALID)
		cts.sync_period = tp->tinfo.curr.period;
	if (spi_valid & CCB_TRANS_SYNC_OFFSET_VALID)
		cts.sync_offset = tp->tinfo.curr.offset;
	if (spi_valid & CCB_TRANS_BUS_WIDTH_VALID)
		cts.bus_width   = tp->tinfo.curr.width;
#endif
	xpt_setup_ccb(&cts.ccb_h, path, /*priority*/1);
	xpt_async(AC_TRANSFER_NEG, path, &cts);
	xpt_free_path(path);
}

/*
 *  Our interrupt handler.
 */
static void sym_intr(void *arg)
{
	if (DEBUG_FLAGS & DEBUG_TINY) printf ("[");
	sym_interrupt((hcb_p) arg);
	if (DEBUG_FLAGS & DEBUG_TINY) printf ("]");
	return;
}

/*
 *  Our polling handler.
 */
static void sym_poll(struct cam_sim *sim)
{
	int s = splcam();
	sym_intr(cam_sim_softc(sim));  
	splx(s);
}

/*
 *  Choose the more appropriate CAM status if 
 *  the IO encountered an extended error.
 */
static int sym_xerr_cam_status(int cam_status, int x_status)
{
	if (x_status) {
		if	(x_status & XE_PARITY_ERR)
			cam_status = CAM_UNCOR_PARITY;
		else if	(x_status &(XE_EXTRA_DATA|XE_SODL_UNRUN|XE_SWIDE_OVRUN))
			cam_status = CAM_DATA_RUN_ERR;
		else if	(x_status & XE_BAD_PHASE)
			cam_status = CAM_REQ_CMP_ERR;
		else
			cam_status = CAM_REQ_CMP_ERR;
	}
	return cam_status;
}

/*
 *  Build CAM result for a failed or auto-sensed IO.
 */
void sym_set_cam_result_error(hcb_p np, ccb_p cp, int resid)
{
	struct ccb_scsiio *csio = &cp->cam_ccb->csio;
	u_int cam_status;

	csio->resid = resid;
	csio->sense_resid = 0;

	if (cp->host_flags & HF_SENSE) {		/* Auto sense     */
		csio->scsi_status = cp->sv_scsi_status;	/* Restore status */
		csio->sense_resid = csio->resid;	/* Swap residuals */
		csio->resid       = cp->sv_resid;
		cp->sv_resid	  = 0;
		if (sym_verbose && cp->sv_xerr_status)
			sym_print_xerr(cp, cp->sv_xerr_status);
		if (cp->host_status == HS_COMPLETE &&
		    cp->ssss_status == S_GOOD &&
		    cp->xerr_status == 0) {
			cam_status = sym_xerr_cam_status(CAM_SCSI_STATUS_ERROR,
							 cp->sv_xerr_status);
			cam_status |= CAM_AUTOSNS_VALID;
			/*
			 *  Bounce back the sense data to user and 
			 *  fix the residual.
			 */
			bzero(&csio->sense_data, csio->sense_len);
			bcopy(cp->sns_bbuf, &csio->sense_data,
			      MIN(csio->sense_len, SYM_SNS_BBUF_LEN));
			csio->sense_resid += csio->sense_len;
			csio->sense_resid -= SYM_SNS_BBUF_LEN;
#if 0
			/*
			 *  If the device reports a UNIT ATTENTION condition 
			 *  due to a RESET condition, we should consider all 
			 *  disconnect CCBs for this unit as aborted.
			 */
			if (1) {
				u_char *p;
				p  = (u_char *) csio->sense_data;
				if (p[0]==0x70 && p[2]==0x6 && p[12]==0x29)
					sym_clear_tasks(np, CAM_REQ_ABORTED,
							cp->target,cp->lun, -1);
			}
#endif
		}
		else
			cam_status = CAM_AUTOSENSE_FAIL;
	}
	else if (cp->host_status == HS_COMPLETE) {	/* Bad SCSI status */
		csio->scsi_status = cp->ssss_status;
		cam_status = CAM_SCSI_STATUS_ERROR;
	}
	else if (cp->host_status == HS_SEL_TIMEOUT)	/* Selection timeout */
		cam_status = CAM_SEL_TIMEOUT;
	else if (cp->host_status == HS_UNEXPECTED)	/* Unexpected BUS FREE*/
		cam_status = CAM_UNEXP_BUSFREE;
	else {						/* Extended error */
		if (sym_verbose) {
			PRINT_ADDR(cp);
			printf ("COMMAND FAILED (%x %x %x).\n",
				cp->host_status, cp->ssss_status,
				cp->xerr_status);
		}
		csio->scsi_status = cp->ssss_status;
		/*
		 *  Set the most appropriate value for CAM status.
		 */
		cam_status = sym_xerr_cam_status(CAM_REQ_CMP_ERR,
						 cp->xerr_status);
	}

	sym_set_cam_status((union ccb *) csio, cam_status);
}

/*
 *  Our SCSI I/O timeout handler.
 */
static void sym_timeout1(void *arg)
{
	union ccb *ccb = (union ccb *) arg;
	hcb_p np = ccb->ccb_h.sym_hcb_ptr;

	/*
	 *  Check that the CAM CCB is still queued.
	 */
	if (!np)
		return;

	switch(ccb->ccb_h.func_code) {
	case XPT_SCSI_IO:
		if (sym_abort_scsiio(np, ccb, 1) == 0)
		ccb->ccb_h.timeout_ch =
			timeout(sym_timeout, (caddr_t) ccb, 10*hz);
		break;
	default:
		break;
	}
}

static void sym_timeout(void *arg)
{
	int s = splcam();
	sym_timeout1(arg);
	splx(s);
}

/*
 *  SIM action entry point.
 */
static void sym_action(struct cam_sim *sim, union ccb *ccb)
{
	int s = splcam();
	sym_action1(sim, ccb);
	splx(s);
}

static void sym_action1(struct cam_sim *sim, union ccb *ccb)
{
	hcb_p	np;
	tcb_p	tp;
	ccb_p	cp;
	int 	tmp;
	struct	ccb_scsiio *csio;
	struct	ccb_hdr  *ccb_h;

	CAM_DEBUG(ccb->ccb_h.path, CAM_DEBUG_TRACE, ("sym_action\n"));

	/*
	 *  Retrieve our controller data structure.
	 */
	np = (hcb_p) cam_sim_softc(sim);

	/*
	 *  The common case is SCSI IO.
	 *  We deal with other ones elsewhere.
	 */
	if (ccb->ccb_h.func_code != XPT_SCSI_IO) {
		sym_action2(sim, ccb);
		return;
	}
	csio  = &ccb->csio;
	ccb_h = &csio->ccb_h;

	/*
	 *  Work around races.
	 */
	if ((ccb_h->status & CAM_STATUS_MASK) != CAM_REQ_INPROG) {
#ifdef DEBUG_2_0_X
printf("XXXX RACE !!! - status = 0x%x\n", ccb_h->status);
#endif
		xpt_done(ccb);
		return;
	}

	/*
	 *  Minimal checkings, so that we will not 
	 *  go outside our tables.
	 */
	if (ccb_h->target_id   == np->myaddr ||
	    ccb_h->target_id   >= SYM_CONF_MAX_TARGET ||
	    ccb_h->target_lun  >= SYM_CONF_MAX_LUN) {
		sym_xpt_done2(np, ccb, CAM_DEV_NOT_THERE);
		return;
        }

	/*
	 *  Retreive the target descriptor.
	 */
	tp = &np->target[ccb_h->target_id];

	/*
	 *  Complete the 1st INQUIRY command with error 
	 *  condition if the device is flagged NOSCAN 
	 *  at BOOT in the NVRAM. This may speed up 
	 *  the boot and maintain coherency with BIOS 
	 *  device numbering. Clearing the flag allows 
	 *  user to rescan skipped devices later.
	 *  We also return error for devices not flagged 
	 *  for SCAN LUNS in the NVRAM since some mono-lun 
	 *  devices behave badly when asked for some non 
	 *  zero LUN. Btw, this is an absolute hack.:-)
	 */
	if (!(ccb_h->flags & CAM_CDB_PHYS) &&
	    (0x12 == ((ccb_h->flags & CAM_CDB_POINTER) ?
		  csio->cdb_io.cdb_ptr[0] : csio->cdb_io.cdb_bytes[0]))) {
		if ((tp->usrflags & SYM_SCAN_BOOT_DISABLED) ||
		    ((tp->usrflags & SYM_SCAN_LUNS_DISABLED) && 
		     ccb_h->target_lun != 0)) {
			tp->usrflags &= ~SYM_SCAN_BOOT_DISABLED;
			sym_xpt_done2(np, ccb, CAM_DEV_NOT_THERE);
			return;
		}
	}

	/*
	 *  Get a control block for this IO.
	 *  Queue the SCSI IO.
	 */
	tmp = 0;
	if ((ccb_h->flags & CAM_TAG_ACTION_VALID) != 0)
		tmp = csio->tag_action;
	cp = sym_get_ccb(np, ccb_h->target_id, ccb_h->target_lun, tmp);
	if (cp)
		(void) sym_queue_scsiio(np, csio, cp);
	else
		sym_xpt_done2(np, ccb, CAM_RESRC_UNAVAIL);
}

/*
 *  Setup buffers and pointers that address the CDB.
 *  I bet, physical CDBs will never be used on the planet, 
 *  since they can be bounced without significant overhead.
 */
static int sym_setup_cdb(hcb_p np, struct ccb_scsiio *csio, ccb_p cp)
{
	struct ccb_hdr *ccb_h;
	u32	cmd_ba;
	int	cmd_len;
	
	ccb_h = &csio->ccb_h;

	/*
	 *  CDB is 16 bytes max.
	 */
	if (csio->cdb_len > sizeof(cp->cdb_buf)) {
		sym_set_cam_status(cp->cam_ccb, CAM_REQ_INVALID);
		return -1;
	}
	cmd_len = csio->cdb_len;

	if (ccb_h->flags & CAM_CDB_POINTER) {
		/* CDB is a pointer */
		if (!(ccb_h->flags & CAM_CDB_PHYS)) {
			/* CDB pointer is virtual */
			bcopy(csio->cdb_io.cdb_ptr, cp->cdb_buf, cmd_len);
			cmd_ba = CCB_BA (cp, cdb_buf[0]);
		} else {
			/* CDB pointer is physical */
#if 0
			cmd_ba = ((u32)csio->cdb_io.cdb_ptr) & 0xffffffff;
#else
			sym_set_cam_status(cp->cam_ccb, CAM_REQ_INVALID);
			return -1;
#endif
		}
	} else {
		/* CDB is in the CAM ccb (buffer) */
		bcopy(csio->cdb_io.cdb_bytes, cp->cdb_buf, cmd_len);
		cmd_ba = CCB_BA (cp, cdb_buf[0]);
	}

	cp->phys.cmd.addr	= cpu_to_scr(cmd_ba);
	cp->phys.cmd.size	= cpu_to_scr(cmd_len);

	return 0;
}

#ifdef	FreeBSD_Bus_Dma_Abstraction
/*
 *  Call back routine for the DMA map service.
 *  If bounce buffers are used (why ?), we may sleep and then 
 *  be called there in another context.
 */
static void
sym_execute_ccb(void *arg, bus_dma_segment_t *psegs, int nsegs, int error)
{
	ccb_p	cp;
	hcb_p	np;
	union	ccb *ccb;
	int	s;

	s = splcam();

	cp  = (ccb_p) arg;
	ccb = cp->cam_ccb;
	np  = (hcb_p) cp->s.arg;

	/*
	 *  Deal with weird races.
	 */
	if (sym_get_cam_status(ccb) != CAM_REQ_INPROG)
		goto out_abort;

	/*
	 *  Deal with weird errors.
	 */
	if (error) {
		cp->s.dmamapped = 0;
		sym_set_cam_status(cp->cam_ccb, CAM_REQ_ABORTED);
		goto out_abort;
	}

	/*
	 *  Build the data descriptor for the chip.
	 */
	if (nsegs) {
		int retv;
		/* 896 rev 1 requires to be careful about boundaries */
		if (np->device_id == PCI_ID_SYM53C896 && np->revision_id <= 1)
			retv = sym_scatter_sg_physical(np, cp, psegs, nsegs);
		else
			retv = sym_fast_scatter_sg_physical(np,cp, psegs,nsegs);
		if (retv < 0) {
			sym_set_cam_status(cp->cam_ccb, CAM_REQ_TOO_BIG);
			goto out_abort;
		}
	}

	/*
	 *  Synchronize the DMA map only if we have 
	 *  actually mapped the data.
	 */
	if (cp->s.dmamapped) {
		bus_dmamap_sync(np->s.data_dmat, cp->s.dmamap,
			(bus_dmasync_op_t)(cp->s.dmamapped == SYM_DMA_READ ? 
				BUS_DMASYNC_PREREAD : BUS_DMASYNC_PREWRITE));
	}

	/*
	 *  Set host status to busy state.
	 *  May have been set back to HS_WAIT to avoid a race.
	 */
	cp->host_status	= cp->nego_status ? HS_NEGOTIATE : HS_BUSY;

	/*
	 *  Set data pointers.
	 */
	sym_setup_data_pointers(np, cp,  (ccb->ccb_h.flags & CAM_DIR_MASK));

	/*
	 *  Enqueue this IO in our pending queue.
	 */
	sym_enqueue_cam_ccb(np, ccb);

	/*
	 *  When `#ifed 1', the code below makes the driver 
	 *  panic on the first attempt to write to a SCSI device.
	 *  It is the first test we want to do after a driver 
	 *  change that does not seem obviously safe. :)
	 */
#if 0
	switch (cp->cdb_buf[0]) {
	case 0x0A: case 0x2A: case 0xAA:
		panic("XXXXXXXXXXXXX WRITE NOT YET ALLOWED XXXXXXXXXXXXXX\n");
		MDELAY(10000);
		break;
	default:
		break;
	}
#endif
	/*
	 *  Activate this job.
	 */
	sym_put_start_queue(np, cp);
out:
	splx(s);
	return;
out_abort:
	sym_free_ccb(np, cp);
	sym_xpt_done(np, ccb);
	goto out;
}

/*
 *  How complex it gets to deal with the data in CAM.
 *  The Bus Dma stuff makes things still more complex.
 */
int sym_setup_data_and_start(hcb_p np, struct ccb_scsiio *csio, ccb_p cp)
{
	struct ccb_hdr *ccb_h;
	int dir, retv;
	
	ccb_h = &csio->ccb_h;

	/*
	 *  Set up the CDB.
	 */
	if (sym_setup_cdb(np, csio, cp))
		goto out_abort;

	/*
	 *  Now deal with the data.
	 */
	cp->data_len = csio->dxfer_len;
	cp->s.arg    = np;

	/*
	 *  No direction means no data.
	 */
	dir = (ccb_h->flags & CAM_DIR_MASK);
	if (dir == CAM_DIR_NONE) {
		sym_execute_ccb(cp, NULL, 0, 0);
		return 0;
	}

	if (!(ccb_h->flags & CAM_SCATTER_VALID)) {
		/* Single buffer */
		if (!(ccb_h->flags & CAM_DATA_PHYS)) {
			/* Buffer is virtual */
			int s;

			cp->s.dmamapped = (dir == CAM_DIR_IN) ? 
						SYM_DMA_READ : SYM_DMA_WRITE;
			s = splsoftvm();
			retv = bus_dmamap_load(np->s.data_dmat, cp->s.dmamap,
					       csio->data_ptr, csio->dxfer_len,
					       sym_execute_ccb, cp, 0);
			if (retv == EINPROGRESS) {
				cp->host_status	= HS_WAIT;
				xpt_freeze_simq(np->s.sim, 1);
				csio->ccb_h.status |= CAM_RELEASE_SIMQ;
			}
			splx(s);
		} else {
			/* Buffer is physical */
			struct bus_dma_segment seg;
#if 0
			seg.ds_addr = (bus_addr_t) csio->data_ptr;
#else
			seg.ds_addr = (bus_addr_t) ((intptr_t) csio->data_ptr);
#endif
			sym_execute_ccb(cp, &seg, 1, 0);
		}
	} else {
		/* Scatter/gather list */
		struct bus_dma_segment *segs;

		if ((ccb_h->flags & CAM_SG_LIST_PHYS) != 0) {
			/* The SG list pointer is physical */
			sym_set_cam_status(cp->cam_ccb, CAM_REQ_INVALID);
			goto out_abort;
		}

		if (!(ccb_h->flags & CAM_DATA_PHYS)) {
			/* SG buffer pointers are virtual */
			sym_set_cam_status(cp->cam_ccb, CAM_REQ_INVALID);
			goto out_abort;
		}

		/* SG buffer pointers are physical */
		segs  = (struct bus_dma_segment *)csio->data_ptr;
		sym_execute_ccb(cp, segs, csio->sglist_cnt, 0);
	}
	return 0;
out_abort:
	sym_free_ccb(np, cp);
	sym_xpt_done(np, (union ccb *) csio);
	return 0;
}

/*
 *  Move the scatter list to our data block.
 */
static int 
sym_fast_scatter_sg_physical(hcb_p np, ccb_p cp, 
			     bus_dma_segment_t *psegs, int nsegs)
{
	struct sym_tblmove *data;
	bus_dma_segment_t *psegs2;

	if (nsegs > SYM_CONF_MAX_SG)
		return -1;

	data   = &cp->phys.data[SYM_CONF_MAX_SG-1];
	psegs2 = &psegs[nsegs-1];
	cp->segments = nsegs;

	while (1) {
		sym_build_sge(np, data, psegs2->ds_addr, psegs2->ds_len);
		if (DEBUG_FLAGS & DEBUG_SCATTER) {
			printf ("%s scatter: paddr=%lx len=%ld\n",
				sym_name(np), (long) psegs2->ds_addr,
				(long) psegs2->ds_len);
		}
		if (psegs2 != psegs) {
			--data;
			--psegs2;
			continue;
		}
		break;
	}
	return 0;
}

#else	/* ! FreeBSD_Bus_Dma_Abstraction */

/*
 *  How complex it gets to deal with the data in CAM.
 *  Variant without the Bus Dma Abstraction option.
 */
int sym_setup_data_and_start(hcb_p np, struct ccb_scsiio *csio, ccb_p cp)
{
	struct ccb_hdr *ccb_h;
	int dir, retv;
	
	ccb_h = &csio->ccb_h;

	/*
	 *  Set up the CDB.
	 */
	if (sym_setup_cdb(np, csio, cp))
		goto out_abort;

	/*
	 *  Now deal with the data.
	 */
	cp->data_len = 0;
	cp->segments = 0;

	/*
	 *  No direction means no data.
	 */
	dir = (ccb_h->flags & CAM_DIR_MASK);
	if (dir == CAM_DIR_NONE)
		goto end_scatter;

	if (!(ccb_h->flags & CAM_SCATTER_VALID)) {
		/* Single buffer */
		if (!(ccb_h->flags & CAM_DATA_PHYS)) {
			/* Buffer is virtual */
			retv = sym_scatter_virtual(np, cp,
						(vm_offset_t) csio->data_ptr, 
						(vm_size_t) csio->dxfer_len);
		} else {
			/* Buffer is physical */
			retv = sym_scatter_physical(np, cp,
						(vm_offset_t) csio->data_ptr, 
						(vm_size_t) csio->dxfer_len);
		}
	} else {
		/* Scatter/gather list */
		int nsegs;
		struct bus_dma_segment *segs;
		segs  = (struct bus_dma_segment *)csio->data_ptr;
		nsegs = csio->sglist_cnt;

		if ((ccb_h->flags & CAM_SG_LIST_PHYS) != 0) {
			/* The SG list pointer is physical */
			sym_set_cam_status(cp->cam_ccb, CAM_REQ_INVALID);
			goto out_abort;
		}
		if (!(ccb_h->flags & CAM_DATA_PHYS)) {
			/* SG buffer pointers are virtual */
			retv = sym_scatter_sg_virtual(np, cp, segs, nsegs); 
		} else {
			/* SG buffer pointers are physical */
			retv = sym_scatter_sg_physical(np, cp, segs, nsegs);
		}
	}
	if (retv < 0) {
		sym_set_cam_status(cp->cam_ccb, CAM_REQ_TOO_BIG);
		goto out_abort;
	}

end_scatter:
	/*
	 *  Set data pointers.
	 */
	sym_setup_data_pointers(np, cp, dir);

	/*
	 *  Enqueue this IO in our pending queue.
	 */
	sym_enqueue_cam_ccb(np, (union ccb *) csio);

	/*
	 *  Activate this job.
	 */
	sym_put_start_queue(np, cp);

	/*
	 *  Command is successfully queued.
	 */
	return 0;
out_abort:
	sym_free_ccb(np, cp);
	sym_xpt_done(np, (union ccb *) csio);
	return 0;
}

/*
 *  Scatter a virtual buffer into bus addressable chunks.
 */
static int
sym_scatter_virtual(hcb_p np, ccb_p cp, vm_offset_t vaddr, vm_size_t len)
{
	u_long	pe, pn;
	u_long	n, k; 
	int s;

	cp->data_len += len;

	pe = vaddr + len;
	n  = len;
	s  = SYM_CONF_MAX_SG - 1 - cp->segments;

	while (n && s >= 0) {
		pn = (pe - 1) & ~PAGE_MASK;
		k = pe - pn;
		if (k > n) {
			k  = n;
			pn = pe - n;
		}
		if (DEBUG_FLAGS & DEBUG_SCATTER) {
			printf ("%s scatter: va=%lx pa=%lx siz=%ld\n",
				sym_name(np), pn, (u_long) vtobus(pn), k);
		}
		cp->phys.data[s].addr = cpu_to_scr(vtobus(pn));
		cp->phys.data[s].size = cpu_to_scr(k);
		pe = pn;
		n -= k;
		--s;
	}
	cp->segments = SYM_CONF_MAX_SG - 1 - s;

	return n ? -1 : 0;
}

/*
 *  Scatter a SG list with virtual addresses into bus addressable chunks.
 */
static int
sym_scatter_sg_virtual(hcb_p np, ccb_p cp, bus_dma_segment_t *psegs, int nsegs)
{
	int i, retv = 0;

	for (i = nsegs - 1 ;  i >= 0 ; --i) {
		retv = sym_scatter_virtual(np, cp,
					   psegs[i].ds_addr, psegs[i].ds_len);
		if (retv < 0)
			break;
	}
	return retv;
}

/*
 *  Scatter a physical buffer into bus addressable chunks.
 */
static int
sym_scatter_physical(hcb_p np, ccb_p cp, vm_offset_t paddr, vm_size_t len)
{
	struct bus_dma_segment seg;

	seg.ds_addr = paddr;
	seg.ds_len  = len;
	return sym_scatter_sg_physical(np, cp, &seg, 1);
}

#endif	/* FreeBSD_Bus_Dma_Abstraction */

/*
 *  Scatter a SG list with physical addresses into bus addressable chunks.
 *  We need to ensure 16MB boundaries not to be crossed during DMA of 
 *  each segment, due to some chips being flawed.
 */
#define BOUND_MASK ((1UL<<24)-1)
static int
sym_scatter_sg_physical(hcb_p np, ccb_p cp, bus_dma_segment_t *psegs, int nsegs)
{
	bus_addr_t ps, pe, pn;
	u_long	k; 
	int s, t;

#ifndef	FreeBSD_Bus_Dma_Abstraction
	s  = SYM_CONF_MAX_SG - 1 - cp->segments;
#else
	s  = SYM_CONF_MAX_SG - 1;
#endif
	t  = nsegs - 1;
	ps = psegs[t].ds_addr;
	pe = ps + psegs[t].ds_len;

	while (s >= 0) {
		pn = (pe - 1) & ~BOUND_MASK;
		if (pn <= ps)
			pn = ps;
		k = pe - pn;
		if (DEBUG_FLAGS & DEBUG_SCATTER) {
			printf ("%s scatter: paddr=%lx len=%ld\n",
				sym_name(np), (u_long) pn, k);
		}
		sym_build_sge(np, &cp->phys.data[s], pn, k);
#ifndef	FreeBSD_Bus_Dma_Abstraction
		cp->data_len += k;
#endif
		--s;
		if (pn == ps) {
			if (--t < 0)
				break;
			ps = psegs[t].ds_addr;
			pe = ps + psegs[t].ds_len;
		}
		else
			pe = pn;
	}

	cp->segments = SYM_CONF_MAX_SG - 1 - s;

	return t >= 0 ? -1 : 0;
}
#undef BOUND_MASK

/*
 *  SIM action for non performance critical stuff.
 */
static void sym_action2(struct cam_sim *sim, union ccb *ccb)
{
	hcb_p	np;
	tcb_p	tp;
	lcb_p	lp;
	struct	ccb_hdr  *ccb_h;

	/*
	 *  Retrieve our controller data structure.
	 */
	np = (hcb_p) cam_sim_softc(sim);

	ccb_h = &ccb->ccb_h;

	switch (ccb_h->func_code) {
	case XPT_SET_TRAN_SETTINGS:
	{
		struct ccb_trans_settings *cts;

		cts  = &ccb->cts;
		tp = &np->target[ccb_h->target_id];

		/*
		 *  If the LCB is not yet available and the LUN
		 *  has been probed ok, try to allocate the LCB.
		 */
		lp = sym_lp(np, tp, ccb_h->target_lun);
		if (!lp && sym_is_bit(tp->lun_map, ccb_h->target_lun))
			lp = sym_alloc_lcb(np, ccb_h->target_id,
			                       ccb_h->target_lun);

		/*
		 *  Update SPI transport settings in TARGET control block.
		 *  Update SCSI device settings in LUN control block.
		 */
#ifdef	FreeBSD_New_Tran_Settings
		if (cts->type == CTS_TYPE_CURRENT_SETTINGS) {
#else
		if ((cts->flags & CCB_TRANS_CURRENT_SETTINGS) != 0) {
#endif
			sym_update_trans(np, tp, &tp->tinfo.goal, cts);
			if (lp)
				sym_update_dflags(np, &lp->curr_flags, cts);
		}
#ifdef	FreeBSD_New_Tran_Settings
		if (cts->type == CTS_TYPE_USER_SETTINGS) {
#else
		if ((cts->flags & CCB_TRANS_USER_SETTINGS) != 0) {
#endif
			sym_update_trans(np, tp, &tp->tinfo.user, cts);
			if (lp)
				sym_update_dflags(np, &lp->user_flags, cts);
		}

		sym_xpt_done2(np, ccb, CAM_REQ_CMP);
		break;
	}
	case XPT_GET_TRAN_SETTINGS:
	{
		struct ccb_trans_settings *cts;
		struct sym_trans *tip;
		u_char dflags;

		cts = &ccb->cts;
		tp = &np->target[ccb_h->target_id];
		lp = sym_lp(np, tp, ccb_h->target_lun);

#ifdef	FreeBSD_New_Tran_Settings
#define	cts__scsi (&cts->proto_specific.scsi)
#define	cts__spi  (&cts->xport_specific.spi)
		if (cts->type == CTS_TYPE_CURRENT_SETTINGS) {
			tip = &tp->tinfo.curr;
			dflags = lp ? lp->curr_flags : 0;
		}
		else {
			tip = &tp->tinfo.user;
			dflags = lp ? lp->user_flags : tp->usrflags;
		}

		cts->protocol  = PROTO_SCSI;
		cts->transport = XPORT_SPI;
		cts->protocol_version  = tip->scsi_version;
		cts->transport_version = tip->spi_version;
		
		cts__spi->sync_period = tip->period;
		cts__spi->sync_offset = tip->offset;
		cts__spi->bus_width   = tip->width;
		cts__spi->ppr_options = tip->options;

		cts__spi->valid = CTS_SPI_VALID_SYNC_RATE
		                | CTS_SPI_VALID_SYNC_OFFSET
		                | CTS_SPI_VALID_BUS_WIDTH
		                | CTS_SPI_VALID_PPR_OPTIONS;
 
		cts__spi->flags &= ~CTS_SPI_FLAGS_DISC_ENB;
		if (dflags & SYM_DISC_ENABLED)
			cts__spi->flags |= CTS_SPI_FLAGS_DISC_ENB;
		cts__spi->valid |= CTS_SPI_VALID_DISC;

		cts__scsi->flags &= ~CTS_SCSI_FLAGS_TAG_ENB;
		if (dflags & SYM_TAGS_ENABLED)
			cts__scsi->flags |= CTS_SCSI_FLAGS_TAG_ENB;
		cts__scsi->valid |= CTS_SCSI_VALID_TQ;
#undef	cts__spi
#undef	cts__scsi
#else
		if ((cts->flags & CCB_TRANS_CURRENT_SETTINGS) != 0) {
			tip = &tp->tinfo.curr;
			dflags = lp ? lp->curr_flags : 0;
		}
		else {
			tip = &tp->tinfo.user;
			dflags = lp ? lp->user_flags : tp->usrflags;
		}
		
		cts->sync_period = tip->period;
		cts->sync_offset = tip->offset;
		cts->bus_width   = tip->width;

		cts->valid = CCB_TRANS_SYNC_RATE_VALID
			   | CCB_TRANS_SYNC_OFFSET_VALID
			   | CCB_TRANS_BUS_WIDTH_VALID;

		cts->flags &= ~(CCB_TRANS_DISC_ENB|CCB_TRANS_TAG_ENB);

		if (dflags & SYM_DISC_ENABLED)
			cts->flags |= CCB_TRANS_DISC_ENB;

		if (dflags & SYM_TAGS_ENABLED)
			cts->flags |= CCB_TRANS_TAG_ENB;

		cts->valid |= CCB_TRANS_DISC_VALID;
		cts->valid |= CCB_TRANS_TQ_VALID;
#endif
		sym_xpt_done2(np, ccb, CAM_REQ_CMP);
		break;
	}
	case XPT_CALC_GEOMETRY:
	{
		struct ccb_calc_geometry *ccg;
		u32 size_mb;
		u32 secs_per_cylinder;
		int extended;

		/*
		 *  Silly DOS geometry.  
		 */
		ccg = &ccb->ccg;
		size_mb = ccg->volume_size
			/ ((1024L * 1024L) / ccg->block_size);
		extended = 1;
		
		if (size_mb > 1024 && extended) {
			ccg->heads = 255;
			ccg->secs_per_track = 63;
		} else {
			ccg->heads = 64;
			ccg->secs_per_track = 32;
		}
		secs_per_cylinder = ccg->heads * ccg->secs_per_track;
		ccg->cylinders = ccg->volume_size / secs_per_cylinder;
		sym_xpt_done2(np, ccb, CAM_REQ_CMP);
		break;
	}
	case XPT_PATH_INQ:
	{
		struct ccb_pathinq *cpi = &ccb->cpi;
		cpi->version_num = 1;
		cpi->hba_inquiry = PI_MDP_ABLE|PI_SDTR_ABLE|PI_TAG_ABLE;
		if ((np->features & FE_WIDE) != 0)
			cpi->hba_inquiry |= PI_WIDE_16;
		cpi->target_sprt = 0;
		cpi->hba_misc = 0;
		if (np->usrflags & SYM_SCAN_TARGETS_HILO)
			cpi->hba_misc |= PIM_SCANHILO;
		if (np->usrflags & SYM_AVOID_BUS_RESET)
			cpi->hba_misc |= PIM_NOBUSRESET;
		cpi->hba_eng_cnt = 0;
		cpi->max_target = (np->features & FE_WIDE) ? 15 : 7;
		/* Semantic problem:)LUN number max = max number of LUNs - 1 */
		cpi->max_lun = SYM_CONF_MAX_LUN-1;
		if (SYM_SETUP_MAX_LUN < SYM_CONF_MAX_LUN)
			cpi->max_lun = SYM_SETUP_MAX_LUN-1;
		cpi->bus_id = cam_sim_bus(sim);
		cpi->initiator_id = np->myaddr;
		cpi->base_transfer_speed = 3300;
		strncpy(cpi->sim_vid, "FreeBSD", SIM_IDLEN);
		strncpy(cpi->hba_vid, "Symbios", HBA_IDLEN);
		strncpy(cpi->dev_name, cam_sim_name(sim), DEV_IDLEN);
		cpi->unit_number = cam_sim_unit(sim);
#ifdef	FreeBSD_New_Tran_Settings
		cpi->protocol = PROTO_SCSI;
		cpi->protocol_version = SCSI_REV_2;
		cpi->transport = XPORT_SPI;
		cpi->transport_version = 2;
		cpi->xport_specific.spi.ppr_options = SID_SPI_CLOCK_ST;
		if (np->features & FE_ULTRA3) {
			cpi->transport_version = 3;
			cpi->xport_specific.spi.ppr_options =
			    SID_SPI_CLOCK_DT_ST;
		}
#endif
		sym_xpt_done2(np, ccb, CAM_REQ_CMP);
		break;
	}
	case XPT_ABORT:
	{
		union ccb *abort_ccb = ccb->cab.abort_ccb;
		switch(abort_ccb->ccb_h.func_code) {
		case XPT_SCSI_IO:
			if (sym_abort_scsiio(np, abort_ccb, 0) == 0) {
				sym_xpt_done2(np, ccb, CAM_REQ_CMP);
				break;
			}
		default:
			sym_xpt_done2(np, ccb, CAM_UA_ABORT);
			break;
		}
		break;
	}
	case XPT_RESET_DEV:
	{
		if (ccb_h->target_id   == np->myaddr ||
		    ccb_h->target_id   >= SYM_CONF_MAX_TARGET) {
			sym_xpt_done2(np, ccb, CAM_DEV_NOT_THERE);
			break;
		}
		sym_reset_scsi_target(np, ccb_h->target_id);
		sym_xpt_done2(np, ccb, CAM_REQ_CMP);
		break;
	}
	case XPT_RESET_BUS:
	{
		sym_reset_scsi_bus(np, 0);
		if (sym_verbose) {
			xpt_print_path(np->s.path);
			printf("SCSI BUS reset delivered.\n");
		}
		sym_start_up (np, 1);
		sym_xpt_done2(np, ccb, CAM_REQ_CMP);
		break;
	}
	case XPT_ACCEPT_TARGET_IO:
	case XPT_CONT_TARGET_IO:
	case XPT_EN_LUN:
	case XPT_NOTIFY_ACK:
	case XPT_IMMED_NOTIFY:
	case XPT_TERM_IO:
	default:
		sym_xpt_done2(np, ccb, CAM_REQ_INVALID);
		break;
	}
}

/*
 *  Asynchronous notification handler.
 */
static void
sym_async(void *cb_arg, u32 code, struct cam_path *path, void *arg)
{
	hcb_p np;
	struct cam_sim *sim;
	u_int tn;
	tcb_p tp;
	int s;

	s = splcam();

	sim = (struct cam_sim *) cb_arg;
	np  = (hcb_p) cam_sim_softc(sim);

	switch (code) {
	case AC_LOST_DEVICE:
		tn = xpt_path_target_id(path);
		if (tn >= SYM_CONF_MAX_TARGET)
			break;

		tp = &np->target[tn];

		tp->to_reset  = 0;
		tp->head.sval = 0;
		tp->head.wval = np->rv_scntl3;
		tp->head.uval = 0;

		tp->tinfo.curr.scsi_version = 2;
		tp->tinfo.curr.spi_version  = 2;
		tp->tinfo.curr.period  = tp->tinfo.goal.period = 0;
		tp->tinfo.curr.offset  = tp->tinfo.goal.offset = 0;
		tp->tinfo.curr.width   = tp->tinfo.goal.width  = BUS_8_BIT;
		tp->tinfo.curr.options = tp->tinfo.goal.options = 0;

		break;
	default:
		break;
	}

	splx(s);
}

/*
 *  Update transfer settings of a target.
 */
static void sym_update_trans(hcb_p np, tcb_p tp, struct sym_trans *tip,
			    struct ccb_trans_settings *cts)
{
	/*
	 *  Update the infos.
	 */
#ifdef	FreeBSD_New_Tran_Settings
#define cts__spi (&cts->xport_specific.spi)
	if ((cts__spi->valid & CTS_SPI_VALID_BUS_WIDTH) != 0)
		tip->width = cts__spi->bus_width;
	if ((cts__spi->valid & CTS_SPI_VALID_SYNC_OFFSET) != 0)
		tip->offset = cts__spi->sync_offset;
	if ((cts__spi->valid & CTS_SPI_VALID_SYNC_RATE) != 0)
		tip->period = cts__spi->sync_period;
	if ((cts__spi->valid & CTS_SPI_VALID_PPR_OPTIONS) != 0)
		tip->options = (cts__spi->ppr_options & PPR_OPT_DT);
	if (cts->protocol_version != PROTO_VERSION_UNSPECIFIED &&
	    cts->protocol_version != PROTO_VERSION_UNKNOWN)
		tip->scsi_version = cts->protocol_version;
	if (cts->transport_version != XPORT_VERSION_UNSPECIFIED &&
	    cts->transport_version != XPORT_VERSION_UNKNOWN)
		tip->spi_version = cts->transport_version;
#undef cts__spi
#else
	if ((cts->valid & CCB_TRANS_BUS_WIDTH_VALID) != 0)
		tip->width = cts->bus_width;
	if ((cts->valid & CCB_TRANS_SYNC_OFFSET_VALID) != 0)
		tip->offset = cts->sync_offset;
	if ((cts->valid & CCB_TRANS_SYNC_RATE_VALID) != 0)
		tip->period = cts->sync_period;
#endif
	/*
	 *  Scale against driver configuration limits.
	 */
	if (tip->width  > SYM_SETUP_MAX_WIDE) tip->width  = SYM_SETUP_MAX_WIDE;
	if (tip->offset > SYM_SETUP_MAX_OFFS) tip->offset = SYM_SETUP_MAX_OFFS;
	if (tip->period < SYM_SETUP_MIN_SYNC) tip->period = SYM_SETUP_MIN_SYNC;

	/*
	 *  Scale against actual controller BUS width.
	 */
	if (tip->width > np->maxwide)
		tip->width  = np->maxwide;

#ifdef	FreeBSD_New_Tran_Settings
	/*
	 *  Only accept DT if controller supports and SYNC/WIDE asked.
	 */
	if (!((np->features & (FE_C10|FE_ULTRA3)) == (FE_C10|FE_ULTRA3)) ||
	    !(tip->width == BUS_16_BIT && tip->offset)) {
		tip->options &= ~PPR_OPT_DT;
	}
#else
	/*
	 *  For now, only assume DT if period <= 9, BUS 16 and offset != 0.
	 */
	tip->options = 0;
	if ((np->features & (FE_C10|FE_ULTRA3)) == (FE_C10|FE_ULTRA3) &&
	    tip->period <= 9 && tip->width == BUS_16_BIT && tip->offset) {
		tip->options |= PPR_OPT_DT;
	}
#endif

	/*
	 *  Scale period factor and offset against controller limits.
	 */
	if (tip->options & PPR_OPT_DT) {
		if (tip->period < np->minsync_dt)
			tip->period = np->minsync_dt;
		if (tip->period > np->maxsync_dt)
			tip->period = np->maxsync_dt;
		if (tip->offset > np->maxoffs_dt)
			tip->offset = np->maxoffs_dt;
	}
	else {
		if (tip->period < np->minsync)
			tip->period = np->minsync;
		if (tip->period > np->maxsync)
			tip->period = np->maxsync;
		if (tip->offset > np->maxoffs)
			tip->offset = np->maxoffs;
	}
}

/*
 *  Update flags for a device (logical unit).
 */
static void 
sym_update_dflags(hcb_p np, u_char *flags, struct ccb_trans_settings *cts)
{
#ifdef	FreeBSD_New_Tran_Settings
#define	cts__scsi (&cts->proto_specific.scsi)
#define	cts__spi  (&cts->xport_specific.spi)
	if ((cts__spi->valid & CTS_SPI_VALID_DISC) != 0) {
		if ((cts__spi->flags & CTS_SPI_FLAGS_DISC_ENB) != 0)
			*flags |= SYM_DISC_ENABLED;
		else
			*flags &= ~SYM_DISC_ENABLED;
	}

	if ((cts__scsi->valid & CTS_SCSI_VALID_TQ) != 0) {
		if ((cts__scsi->flags & CTS_SCSI_FLAGS_TAG_ENB) != 0)
			*flags |= SYM_TAGS_ENABLED;
		else
			*flags &= ~SYM_TAGS_ENABLED;
	}
#undef	cts__spi
#undef	cts__scsi
#else
	if ((cts->valid & CCB_TRANS_DISC_VALID) != 0) {
		if ((cts->flags & CCB_TRANS_DISC_ENB) != 0)
			*flags |= SYM_DISC_ENABLED;
		else
			*flags &= ~SYM_DISC_ENABLED;
	}

	if ((cts->valid & CCB_TRANS_TQ_VALID) != 0) {
		if ((cts->flags & CCB_TRANS_TAG_ENB) != 0)
			*flags |= SYM_TAGS_ENABLED;
		else
			*flags &= ~SYM_TAGS_ENABLED;
	}
#endif
}


/*============= DRIVER INITIALISATION ==================*/

#ifdef FreeBSD_Bus_Io_Abstraction

static device_method_t sym_pci_methods[] = {
	DEVMETHOD(device_probe,	 sym_pci_probe),
	DEVMETHOD(device_attach, sym_pci_attach),
	{ 0, 0 }
};

static driver_t sym_pci_driver = {
	"sym",
	sym_pci_methods,
	sizeof(struct sym_hcb)
};

static devclass_t sym_devclass;

DRIVER_MODULE(sym, pci, sym_pci_driver, sym_devclass, 0, 0);

#else	/* Pre-FreeBSD_Bus_Io_Abstraction */

static u_long sym_unit;

static struct	pci_device sym_pci_driver = {
	"sym",
	sym_pci_probe,
	sym_pci_attach,
	&sym_unit,
	NULL
}; 

#if 	__FreeBSD_version >= 400000
COMPAT_PCI_DRIVER (sym, sym_pci_driver);
#else
DATA_SET (pcidevice_set, sym_pci_driver);
#endif

#endif /* FreeBSD_Bus_Io_Abstraction */

/*
 *  Look up the chip table.
 *
 *  Return a pointer to the chip entry if found, 
 *  zero otherwise.
 */
static struct sym_pci_chip *
#ifdef FreeBSD_Bus_Io_Abstraction
sym_find_pci_chip(device_t dev)
#else
sym_find_pci_chip(pcici_t dev)
#endif
{
	u_short	device_id;
	u_char	revision;

	if (sym_pci_get_vendor(dev) != PCI_VENDOR_NCR)
		return 0;

	device_id = sym_pci_get_device(dev);
	revision  = sym_pci_get_revid(dev);

	return sym_lookup_pci_chip_table(device_id, revision);
}

/*
 *  Check if the chip is owned by RAID.
 *  Signature is 0x52414944 at (end-of-RAM - 16).
 */
static int 
sym_owned_by_raid(sym_device_t dev, struct sym_pci_chip *chip)
{
#ifdef FreeBSD_Bus_Io_Abstraction
	int regs_id;
	u32 tmp;
	struct sym_bus_mem ram;
	int offset;

	if (!(chip->features & (FE_RAM|FE_RAM8K)))
		return 0;
	regs_id = (chip->features & FE_64BIT)? SYM_PCI_RAM64 : SYM_PCI_RAM;
	if (!sym_map_bus_mem(dev, regs_id, &ram))
		return 0;
	offset = (chip->features & FE_RAM8K)? (8192-16) : (4096-16); 
#ifdef FreeBSD_Bus_Io_Abstraction
	tmp = bus_space_read_4(ram.tag, ram.bsh, offset);
#else
	tmp = *((volatile u32 *)(ram.va + offset));
#endif
	sym_unmap_bus_mem(dev, &ram);
	return (tmp == 0x52414944);
#else
	return 0;
#endif
}

/*
 *  Tell upper layer if the chip is supported.
 */
#ifdef FreeBSD_Bus_Io_Abstraction
static int
sym_pci_probe(device_t dev)
{
	struct	sym_pci_chip *chip;

	chip = sym_find_pci_chip(dev);
	if (chip && sym_find_firmware(chip) && !sym_owned_by_raid(dev, chip)) {
		device_set_desc(dev, chip->name);
		return (chip->lp_probe_bit & SYM_SETUP_LP_PROBE_MAP)? -2000 : 0;
	}
	return ENXIO;
}
#else /* Pre-FreeBSD_Bus_Io_Abstraction */
static const char *
sym_pci_probe(pcici_t dev, pcidi_t type)
{
	struct	sym_pci_chip *chip;

	chip = sym_find_pci_chip(dev);
	if (chip && sym_find_firmware(chip) && !sym_owned_by_raid(dev, chip)) {
#if NNCR > 0
	/* Only claim chips we are allowed to take precedence over the ncr */
	if (!(chip->lp_probe_bit & SYM_SETUP_LP_PROBE_MAP))
#else
	if (1)
#endif
		return chip->name;
	}
	return 0;
}
#endif

/*
 *  Attach a sym53c8xx device.
 */
#ifdef FreeBSD_Bus_Io_Abstraction
static int
sym_pci_attach(device_t dev)
#else
static void
sym_pci_attach(pcici_t dev, int unit)
{
	int err = sym_pci_attach2(dev, unit);
	if (err)
		printf("sym: failed to attach unit %d - err=%d.\n", unit, err);
}
static int
sym_pci_attach2(pcici_t dev, int unit)
#endif
{
	struct	sym_pci_chip *chip;
	u_short	command;
	u_char	cachelnsz;
	struct	sym_hcb *np = 0;
	struct	sym_fw *fw = 0;
	int 	i;
#ifdef	FreeBSD_Bus_Dma_Abstraction
	bus_dma_tag_t	bus_dmat;

	/*
	 *  I expected to be told about a parent 
	 *  DMA tag, but didn't find any.
	 */
	bus_dmat = NULL;
#endif

	/*
	 *  Only probed devices should be attached.
	 *  We just enjoy being paranoid. :)
	 */
	chip = sym_find_pci_chip(dev);
	if (chip == NULL || (fw = sym_find_firmware(chip)) == NULL)
		return (ENXIO);

	/*
	 *  Allocate immediately the host control block, 
	 *  since we are only expecting to succeed. :)
	 *  We keep track in the HCB of all the resources that 
	 *  are to be released on error.
	 */
#ifdef	FreeBSD_Bus_Dma_Abstraction
	np = __sym_calloc_dma(bus_dmat, sizeof(*np), "HCB");
	if (np)
		np->bus_dmat = bus_dmat;
	else
		goto attach_failed;
#else
	np = sym_calloc_dma(sizeof(*np), "HCB");
	if (!np)
		goto attach_failed;
#endif

	/*
	 *  Copy some useful infos to the HCB.
	 */
	np->hcb_ba	 = vtobus(np);
	np->verbose	 = bootverbose;
	np->s.device	 = dev;
#ifdef	FreeBSD_Bus_Io_Abstraction
	np->s.unit	 = device_get_unit(dev);
#else
	np->s.unit	 = unit;
#endif
	np->device_id	 = sym_pci_get_device(dev);
	np->revision_id  = sym_pci_get_revid(dev);
	np->features	 = chip->features;
	np->clock_divn	 = chip->nr_divisor;
	np->maxoffs	 = chip->offset_max;
	np->maxburst	 = chip->burst_max;

	/*
	 * Edit its name.
	 */
	snprintf(np->s.inst_name, sizeof(np->s.inst_name), "sym%d", np->s.unit);

	/*
	 *  Allocate a tag for the DMA of user data.
	 */
#ifdef	FreeBSD_Bus_Dma_Abstraction
#if	SYM_CONF_DMA_ADDRESSING_MODE == 0
#define	LowAddr		BUS_SPACE_MAXADDR_32BIT
#define	HighAddr	BUS_SPACE_MAXADDR
#elif	SYM_CONF_DMA_ADDRESSING_MODE == 1
#define	LowAddr		\
	((np->features & FE_DAC) ? 0xffffffffff : BUS_SPACE_MAXADDR_32BIT)
#define	HighAddr	BUS_SPACE_MAXADDR
#define	UseDac		\
	((np->features & FE_DAC) ? 1 : 0)
#endif
	if (bus_dma_tag_create(np->bus_dmat, 1, (1<<24),
				LowAddr, HighAddr,
				NULL, NULL,
				BUS_SPACE_MAXSIZE, SYM_CONF_MAX_SG,
				(1<<24), 0, &np->s.data_dmat)) {
		device_printf(dev, "failed to create DMA tag.\n");
		goto attach_failed;
	}
#if	SYM_CONF_DMA_ADDRESSING_MODE != 0
	np->use_dac = UseDac;
#endif
#undef	LowAddr
#undef	HighAddr
#undef	UseDac
#endif	/* FreeBSD_Bus_Dma_Abstraction */

	/*
	 *  Read and apply some fix-ups to the PCI COMMAND 
	 *  register. We want the chip to be enabled for:
	 *  - BUS mastering
	 *  - PCI parity checking (reporting would also be fine)
	 *  - Write And Invalidate.
	 */
	command = sym_pci_read_config_u16(dev, PCIR_COMMAND);
	command |= PCIM_CMD_BUSMASTEREN;
	command |= PCIM_CMD_PERRESPEN;
	command |= /* PCIM_CMD_SEREN */ 0x00100;
	command |= /* PCIM_CMD_MWIEN */ 0x0010;
	sym_pci_write_config_u16(dev, PCIR_COMMAND, command);

	/*
	 *  Let the device know about the cache line size, 
	 *  if it doesn't yet.
	 */
	cachelnsz = sym_pci_read_config_u8(dev, PCIR_CACHELNSZ);
	if (!cachelnsz) {
#if 0
		cachelnsz = 8;
		sym_pci_write_config_u8(dev, PCIR_CACHELNSZ, cachelnsz);
#else
		printf("%s: CACHE LINE SIZE NOT CONFIGURED.\n", sym_name(np));
#endif
	}

	/*
	 *  Allocate/get/map/retrieve everything that deals with MMIO.
	 */
	if ((command & PCIM_CMD_MEMEN) != 0) {
		np->mmio_ba = sym_map_bus_mem(dev, SYM_PCI_MMIO, &np->s.mmio);
		if (!np->mmio_ba) {
			printf("%s: failed to map MMIO window\n", sym_name(np));
			goto attach_failed;
		}
	}
#ifdef	SYM_CONF_IOMAPPED
	/*
	 *  Allocate/get/map/retrieve everything that deals with IO.
	 */
	if ((command & PCI_COMMAND_IO_ENABLE) != 0) {
		if (!sym_map_bus_io(dev, SYM_PCI_IO, &np->s.io)) {
			printf("%s: failed to map IO window\n", sym_name(np));
			goto attach_failed;
	}
#endif
	/*
	 *  If the chip has RAM.
	 *  Allocate/get/map/retrieve the corresponding resources.
	 */
	if ((np->features & (FE_RAM|FE_RAM8K)) &&
	    (command & PCIM_CMD_MEMEN) != 0) {
		i = (chip->features & FE_64BIT)? SYM_PCI_RAM64 : SYM_PCI_RAM;
		np->ram_ba = sym_map_bus_mem(dev, i, &np->s.ram);
		if (!np->ram_ba) {
			printf("%s: failed to map RAM window\n", sym_name(np));
			goto attach_failed;
		}
	}

	/*
	 *  Allocate the IRQ.
	 */
#ifdef FreeBSD_Bus_Io_Abstraction
	i = 0;
	np->s.irq_res = bus_alloc_resource(dev, SYS_RES_IRQ, &i,
					0, ~0, 1, RF_ACTIVE | RF_SHAREABLE);
	if (!np->s.irq_res) {
		device_printf(dev, "failed to allocate IRQ resource\n");
		goto attach_failed;
	}
#endif

	/*
	 *  Perform O/S independant stuff.
	 */
	if (sym_hcb_attach(np, fw))
		goto attach_failed;

	/*
	 *  Now deal with CAM.
	 *  Hopefully, we will succeed with that one.:)
	 */
	if (!sym_cam_attach(np))
		goto attach_failed;

	/*
	 * Initialyze the CAM CCB pending queue.
	 */
	sym_que_init(&np->s.cam_ccbq);

	/*
	 *  Sigh! we are done.
	 */
	return 0;

	/*
	 *  We have failed.
	 *  We will try to free all the resources we have 
	 *  allocated, but if we are a boot device, this 
	 *  will not help that much.;)
	 */
attach_failed:
	if (np)
		sym_pci_free(np);
	return ENXIO;
}

/*
 *  Free everything that have been allocated for this device.
 */
static void sym_pci_free(hcb_p np)
{
	int s;

	/*
	 *  First free CAM resources.
	 */
	s = splcam();
	sym_cam_free(np);
	splx(s);

	/*
	 *  Free O/S independant resources.
	 */
	sym_hcb_free(np);

	/*
	 *  Now every should be quiet for us to 
	 *  free other resources.
	 */
	sym_unmap_bus_mem(np->s.device, &np->s.ram);
	sym_unmap_bus_mem(np->s.device, &np->s.mmio);
#ifdef	SYM_CONF_IOMAPPED
	sym_unmap_bus_io(np->s.device, &np->s.io);
#endif
#ifdef FreeBSD_Bus_Io_Abstraction
	if (np->s.irq_res)
		bus_release_resource(np->s.device, SYS_RES_IRQ, 
				     0, np->s.irq_res);
#endif
#ifdef	FreeBSD_Bus_Dma_Abstraction
	if (np->s.data_dmat)
		bus_dma_tag_destroy(np->s.data_dmat);
#endif
	sym_mfree_dma(np, sizeof(*np), "HCB");
}

/*
 *  PCI BUS error handler.
 */
void sym_log_bus_error(hcb_p np)
{
	u_short pci_sts;
	pci_sts = sym_pci_read_config_u16(np->s.device, PCIR_STATUS);
	if (pci_sts & 0xf900) {
		sym_pci_write_config_u16(np->s.device, PCIR_STATUS,
		                         pci_sts);
		printf("%s: PCI STATUS = 0x%04x\n",
			sym_name(np), pci_sts & 0xf900);
	}
}

/*
 *  Allocate CAM resources and register a bus to CAM.
 */
static int sym_cam_attach(hcb_p np)
{
	struct cam_devq *devq = 0;
	struct cam_sim *sim = 0;
	struct cam_path *path = 0;
	struct ccb_setasync csa;
	int err, s;

	s = splcam();

	/*
	 *  Establish our interrupt handler.
	 */
#ifdef FreeBSD_Bus_Io_Abstraction
	err = bus_setup_intr(np->s.device, np->s.irq_res, 
#if 	__FreeBSD_version >= 500000
			     INTR_TYPE_CAM | INTR_ENTROPY,
#else
			     INTR_TYPE_CAM,
#endif
			     sym_intr, np, &np->s.intr);
	if (err) {
		device_printf(np->s.device, "bus_setup_intr() failed: %d\n",
			      err);
		goto fail;
	}
#else
	err = 0;
	if (!pci_map_int (np->s.device, sym_intr, np, &cam_imask)) {
		printf("%s: failed to map interrupt\n", sym_name(np));
		goto fail;
	}
#endif

	/*
	 *  Create the device queue for our sym SIM.
	 */
	devq = cam_simq_alloc(SYM_CONF_MAX_START);
	if (!devq)
		goto fail;

	/*
	 *  Construct our SIM entry.
	 */
	sim = cam_sim_alloc(sym_action, sym_poll, "sym", np, np->s.unit,
			    1, SYM_SETUP_MAX_TAG, devq);
	if (!sim)
		goto fail;
	devq = 0;

	if (xpt_bus_register(sim, 0) != CAM_SUCCESS)
		goto fail;
	np->s.sim = sim;
	sim = 0;

	if (xpt_create_path(&path, 0,
			    cam_sim_path(np->s.sim), CAM_TARGET_WILDCARD,
			    CAM_LUN_WILDCARD) != CAM_REQ_CMP) {
		goto fail;
	}
	np->s.path = path;

	/*
	 *  Hmmm... This should be useful, but I donnot want to 
	 *  know about.
	 */
#if 	__FreeBSD_version < 400000
#ifdef	__alpha__
	alpha_register_pci_scsi(sym_pci_get_bus(np->s.device),
				sym_pci_get_slot(np->s.device), np->s.sim);
#endif
#endif

	/*
	 *  Establish our async notification handler.
	 */
	xpt_setup_ccb(&csa.ccb_h, np->s.path, 5);
	csa.ccb_h.func_code = XPT_SASYNC_CB;
	csa.event_enable    = AC_LOST_DEVICE;
	csa.callback	    = sym_async;
	csa.callback_arg    = np->s.sim;
	xpt_action((union ccb *)&csa);

	/*
	 *  Start the chip now, without resetting the BUS, since  
	 *  it seems that this must stay under control of CAM.
	 *  With LVD/SE capable chips and BUS in SE mode, we may 
	 *  get a spurious SMBC interrupt.
	 */
	sym_start_up (np, 0);

	splx(s);
	return 1;
fail:
	if (sim)
		cam_sim_free(sim, FALSE);
	if (devq)
		cam_simq_free(devq);

	sym_cam_free(np);

	splx(s);
	return 0;
}

/*
 *  Free everything that deals with CAM.
 */
static void sym_cam_free(hcb_p np)
{
#ifdef FreeBSD_Bus_Io_Abstraction
	if (np->s.intr)
		bus_teardown_intr(np->s.device, np->s.irq_res, np->s.intr);
#else
	/* pci_unmap_int(np->s.device); */	/* Does nothing */
#endif
	
	if (np->s.sim) {
		xpt_bus_deregister(cam_sim_path(np->s.sim));
		cam_sim_free(np->s.sim, /*free_devq*/ TRUE);
	}
	if (np->s.path)
		xpt_free_path(np->s.path);
}
