/*
 * Device driver for the SYMBIOS/LSILOGIC 53C8XX and 53C1010 family 
 * of PCI-SCSI IO processors.
 *
 * Copyright (C) 1999-2001  Gerard Roudier <groudier@free.fr>
 *
 * This driver is derived from the Linux sym53c8xx driver.
 * Copyright (C) 1998-2000  Gerard Roudier
 *
 * The sym53c8xx driver is derived from the ncr53c8xx driver that had been 
 * a port of the FreeBSD ncr driver to Linux-1.2.13.
 *
 * The original ncr driver has been written for 386bsd and FreeBSD by
 *         Wolfgang Stanglmeier        <wolf@cologne.de>
 *         Stefan Esser                <se@mi.Uni-Koeln.de>
 * Copyright (C) 1994  Wolfgang Stanglmeier
 *
 * Other major contributions:
 *
 * NVRAM detection and reading.
 * Copyright (C) 1997 Richard Waltham <dormouse@farsrobt.demon.co.uk>
 *
 *-----------------------------------------------------------------------------
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sym_glue.h"

/*
 *  Some stupid macros that make NetBSD 1.4 scsipi resemble 
 *  NetBSD 1.5 scsipi. :)
 */

#define _is_flag_set(f, m)	(((f) & (m)) != 0)
#define _set_flag(f, m)		do { (f) |=  (m); } while (0)
#define _clear_flag(f, m)	do { (f) &= ~(m); } while (0)

#if	__NetBSD_Version__ >= 105000000

#define xs_sts_done_isset(x)	 _is_flag_set((x)->xs_status, XS_STS_DONE)
#define xs_sts_done_set(x)	 _set_flag((x)->xs_status, XS_STS_DONE)
#define xs_sts_done_clear(x)	 _clear_flag((x)->xs_status, XS_STS_DONE)
#define xs_ctl_poll_isset(x)	 _is_flag_set((x)->xs_control, XS_CTL_POLL)
#define xs_ctl_data_in_isset(x)  _is_flag_set((x)->xs_control, XS_CTL_DATA_IN)
#define xs_ctl_data_out_isset(x) _is_flag_set((x)->xs_control, XS_CTL_DATA_OUT)
#define xs_ctl_data_uio_isset(x) _is_flag_set((x)->xs_control, XS_CTL_DATA_UIO)
#define xs_ctl_reset_isset(x)	 _is_flag_set((x)->xs_control, XS_CTL_RESET)

#else

#define xs_sts_done_isset(x)	 _is_flag_set((x)->flags, ITSDONE)
#define xs_sts_done_set(x)	 _set_flag((x)->flags, ITSDONE)
#define xs_sts_done_clear(x)	 _clear_flag((x)->flags, ITSDONE)
#define xs_ctl_poll_isset(x)	 _is_flag_set((x)->flags, SCSI_POLL)
#define xs_ctl_data_in_isset(x)  _is_flag_set((x)->flags, SCSI_DATA_IN)
#define xs_ctl_data_out_isset(x) _is_flag_set((x)->flags, SCSI_DATA_OUT)
#define xs_ctl_data_uio_isset(x) _is_flag_set((x)->flags, SCSI_DATA_UIO)
#define xs_ctl_reset_isset(x)	 _is_flag_set((x)->flags, SCSI_RESET)

#endif

/*
 *  How verbose user wants us to be.
 */
#ifndef bootverbose
#define bootverbose 1
#endif

/*
 *  BARs offset.
 */
#define PCIR_MAPS	0x10
#define SYM_PCI_IO	PCIR_MAPS
#define SYM_PCI_MMIO	(PCIR_MAPS + 4)
#define SYM_PCI_RAM	(PCIR_MAPS + 8)
#define SYM_PCI_RAM64	(PCIR_MAPS + 12)

/*
 *  Our softc data structure.
 */
struct sym_data {
	struct device dev;	/* Shoe-horned by the O/S :-) */
	hcb_p np;
};

/*
 *  Some needed prototypes.
 */
static int 
sym_fast_scatter_sg_physical(hcb_p np, ccb_p cp, 
			     bus_dma_segment_t *psegs, int nsegs);
static int 
sym_scatter_sg_physical(hcb_p np, ccb_p cp, 
			bus_dma_segment_t *psegs, int nsegs);
static void sym_pci_free(hcb_p np);
static int sym_cam_attach(hcb_p np);
static void sym_cam_free(hcb_p np);

/*
 *  C wrapper function for a macro that faults otherwise.
 */
void 
_bus_space_write_region_1_(bus_space_tag_t t, bus_space_handle_t h, 
			    bus_size_t o, u8 *p, bus_size_t n)
{
	bus_space_write_region_1(t, h, o, p, n);
}

/*
 *  Insert a delay in micro-seconds and milli-seconds.
 */
void sym_udelay(int us) { delay(us); }
void sym_mdelay(int ms) { while (ms--) delay(1000); }

/*
 *  Complete a pending CAM CCB.
 */
void sym_xpt_done(hcb_p np, struct scsipi_xfer *xp)
{
	xs_sts_done_set(xp);
	if (!xs_ctl_poll_isset(xp))
		scsipi_done(xp);
}

void sym_xpt_done2(hcb_p np, struct scsipi_xfer *xp, int cam_status)
{
	sym_set_cam_status(xp, cam_status);
	sym_xpt_done(np, xp);
}

/*
 *  Print something that identifies the IO.
 */
void sym_print_addr(ccb_p cp)
{
	struct scsipi_xfer *xp = cp->cam_ccb;
	if (xp) scsi_print_addr(xp->sc_link);
}

/*
 *  Tell the SCSI layer about a BUS RESET.
 */
void sym_xpt_async_bus_reset(hcb_p np)
{
	printf("%s: SCSI BUS has been reset.\n", sym_name(np));
}

/*
 *  Tell the SCSI layer about a BUS DEVICE RESET message sent.
 */
void sym_xpt_async_sent_bdr(hcb_p np, int target)
{
	printf("%s: TARGET %d has been reset.\n", sym_name(np), target);
}

/*
 *  Tell the SCSI layer about the new transfer parameters.
 */
void sym_xpt_async_nego_wide(hcb_p np, int target)
{
	if (sym_verbose < 3)
		return;
	sym_announce_transfer_rate(np, target);
}

/*
 *  Choose the more appropriate CAM status if 
 *  the IO encountered an extended error.
 */
static int sym_xerr_cam_status(int cam_status, int x_status)
{
	if (x_status)
		return XS_DRIVER_STUFFUP;
	return cam_status;
}

/*
 *  Build CAM result for a failed or auto-sensed IO.
 */
void sym_set_cam_result_error(hcb_p np, ccb_p cp, int resid)
{
	struct scsipi_xfer *xp = cp->cam_ccb;
	u_int scsi_status;

	scsi_status = cp->ssss_status;
	if (cp->host_flags & HF_SENSE) {
		xp->resid = cp->sv_resid;
		xp->status = cp->sv_scsi_status;
		if (sym_verbose && cp->sv_xerr_status)
			sym_print_xerr(cp, cp->sv_xerr_status);
		if (cp->host_status == HS_COMPLETE &&
		    cp->ssss_status == S_GOOD &&
		    cp->xerr_status == 0) {
			xp->error = XS_SENSE;
			if (0) {
				u_char *p;
				p  = (u_char *) &cp->sns_bbuf[0];
				if (p[0]==0x70 && p[2]==0x6 && p[12]==0x29)
					sym_clear_tasks(np, XS_DRIVER_STUFFUP,
							cp->target,cp->lun, -1);
			}
		}
		else {
			xp->error = XS_DRIVER_STUFFUP;
		}
		bzero(&xp->sense, sizeof(xp->sense));
		bcopy(cp->sns_bbuf, &xp->sense,
		      MIN(sizeof(xp->sense),SYM_SNS_BBUF_LEN));
	}
	else if (cp->host_status == HS_COMPLETE) { 	/* Bad SCSI status */
		xp->status = scsi_status;
		xp->error  = XS_NOERROR;
	}
	else if (cp->host_status == HS_SEL_TIMEOUT)	/* Selection timeout */
		xp->error  = XS_SELTIMEOUT;
	else if (cp->host_status == HS_UNEXPECTED)	/* Unexpected BUS FREE*/
		xp->error  = XS_DRIVER_STUFFUP;
	else {						/* Extended error */
		if (sym_verbose) {
			PRINT_ADDR(cp);
			printf ("COMMAND FAILED (%x %x %x).\n",
				cp->host_status, cp->ssss_status,
				cp->xerr_status);
		}
		xp->error =
			sym_xerr_cam_status(XS_DRIVER_STUFFUP, cp->xerr_status);
	}
}

/*
 *  Our interrupt handler.
 */
static int sym_intr(void *arg)
{
	sym_interrupt((hcb_p) arg);
	return 1;
}

/*
 *  Tune device queuing depth, according to various limits.
 */
static void 
sym_tune_dev_queuing(hcb_p np, int target, int lun, u_short reqtags)
{
	tcb_p	tp = &np->target[target];
	lcb_p	lp = sym_lp(np, tp, lun);
	u_short	oldtags;

	if (!lp)
		return;

	oldtags = lp->s.reqtags;

	if (reqtags > lp->s.openings)
		reqtags = lp->s.openings;

	lp->started_limit = reqtags ? reqtags : 2;
	lp->started_max   = 1;
	lp->s.reqtags     = reqtags;

	if (reqtags != oldtags) {
		printf("%s:%d:%d: "
			"tagged command queuing %s, command queue depth %d.\n",
			sym_name(np), target, lun,
			lp->s.reqtags ? "enabled" : "disabled",
			lp->started_limit);
	}
}

/*
 *  Also called from ioctl ACCEL.
 */
static void
sym_setup_dev_queuing(hcb_p np, tcb_p tp, lcb_p lp, struct scsipi_link *scp)
{
	int target = scp->scsipi_scsi.target;
	int lun    = scp->scsipi_scsi.lun;
	u_short reqtags;

	if (!lp->s.openings)
		lp->s.openings = 1;

	reqtags = 0;
	if (tp->usrtags && (lp->curr_flags & SYM_TAGS_ENABLED) &&
	    sym_is_bit(tp->cmdq_map, lun))
		reqtags = SYM_CONF_MAX_TAG;

	if (reqtags > lp->s.openings) {
		scp->openings += (reqtags - lp->s.openings);
		lp->s.openings = reqtags;
	}
#if 0
printf("XXXX sc_link->openings = %d - reqtags = %d XXXX\n",
	scp->openings, reqtags);
#endif
	sym_tune_dev_queuing(np, target, lun, reqtags);
}

/*
 *  Called on successfull INQUIRY response.
 */
void sym_sniff_inquiry(hcb_p np, struct scsipi_xfer *xp, int resid)
{
	int target = xp->sc_link->scsipi_scsi.target;
	int lun    = xp->sc_link->scsipi_scsi.lun;
	tcb_p	tp = &np->target[target];
	lcb_p	lp;
	int retv;

	retv = __sym_sniff_inquiry(np, target, lun, (u_char *)xp->data, 
				   xp->datalen - resid);
	if (retv < 0)
		return;
	else if (retv)
		sym_update_trans_settings(np, tp);

#if __NetBSD_Version__ >= 105000000
	lp = sym_lp(np, tp, lun);
	if (!lp)
		return;
#else
	lp = sym_alloc_lcb(np, target, lun);
	if (!lp)
		return;
	lp->curr_flags = lp->user_flags;
#endif
	sym_setup_dev_queuing(np, tp, lp, xp->sc_link);
}

/*
 *  PCI BUS error handler.
 */
void sym_log_bus_error(hcb_p np)
{
	u32 pci_csr;
	pci_csr = pci_conf_read(np->s.pa_pc, np->s.pa_tag,
				PCI_COMMAND_STATUS_REG);
	if (pci_csr & 0xf9000000) {
		pci_conf_write(np->s.pa_pc, np->s.pa_tag,
			       PCI_COMMAND_STATUS_REG, pci_csr);
		printf("%s: PCI STATUS = 0x%04x\n",
			sym_name(np), (pci_csr & 0xf9000000) >> 16);
	}
}

/*
 *  Action routine.
 */
static int sym_start_io(struct scsipi_xfer *xp)
{
	hcb_p	np;
	tcb_p	tp;
	lcb_p	lp;
	ccb_p	cp;
	int 	order;
	int	s;
	int	result = COMPLETE;

	s = splbio();

	/*
	 *  Retrieve our controller data structure.
	 */
	np = ((struct sym_data *)xp->sc_link->adapter_softc)->np;

	/*
	 *  Minimal checkings, so that we will not 
	 *  go outside our tables.
	 */
	if (sym_cam_target_id(xp)   == np->myaddr ||
	    sym_cam_target_id(xp)   >= SYM_CONF_MAX_TARGET ||
	    sym_cam_target_lun(xp)  >= SYM_CONF_MAX_LUN) {
		xp->error = XS_DRIVER_STUFFUP;
		goto out;
        }

	/*
	 *  Work around weirdnesses.
	 *  Seems most drivers do the same.
	 */
	if (xs_sts_done_isset(xp)) {
		printf("%s: DONE?\n", sym_name (np));
		xs_sts_done_clear(xp);
	}

	/*
	 *  Reset the BUS if asked for.
	 */
	if (xs_ctl_reset_isset(xp)) {
		sym_reset_scsi_bus(np, 1);
		MDELAY(10);
		goto out;
	}

	/*
	 *  Retreive the target descriptor.
	 */
	tp = &np->target[sym_cam_target_id(xp)];

	/*
	 *  Complete the 1st INQUIRY command with error 
	 *  condition if the device is flagged NOSCAN 
	 *  at BOOT in the NVRAM. This may speed up 
	 *  the boot and maintain coherency with BIOS 
	 *  device numbering. Clearing the flag allows 
	 *  user to rescan skipped devices later.
	 *  We also return error for devices not flagged 
	 *  for SCAN LUNS in the NVRAM since some mono-lun 
	 *  devices behave badly when asked for some non 
	 *  zero LUN. Btw, this is an absolute hack.:-)
	 */
	if ((tp->usrflags & SYM_SCAN_BOOT_DISABLED) ||
	    ((tp->usrflags & SYM_SCAN_LUNS_DISABLED) && 
	     sym_cam_target_lun(xp) != 0)) {
		xp->error = XS_SELTIMEOUT;
		if (xp->cmd->opcode == 0x12)
			tp->usrflags &= ~SYM_SCAN_BOOT_DISABLED;
		goto out;
	}

	/*
	 *  Allocate a CCB with appropriate properties.
	 */
	lp = sym_lp(np, tp, sym_cam_target_lun(xp));
	order = (lp && lp->s.reqtags) ? M_SIMPLE_TAG : 0;

	cp = sym_get_ccb(np, sym_cam_target_id(xp), 
			     sym_cam_target_lun(xp), order);
	if (!cp) {
		result = TRY_AGAIN_LATER;
		goto out;
	}

	/*
	 *  Queue the IO to the controller.
	 */
	result = sym_queue_scsiio(np, xp, cp);
	if (result != SUCCESSFULLY_QUEUED) {
		sym_free_ccb(np, cp);
		goto out;
	}

	/*
	 *  Poll for completion if asked for.
	 */
	if (xs_ctl_poll_isset(xp)) {
		while (!xs_sts_done_isset(xp)) {
			sym_interrupt(np);
			DELAY(1000);
		}
		result = COMPLETE;
	}
out:
	splx(s);
	return result;
}

/*
 *  Setup buffers and pointers that address the CDB.
 */
static int __inline sym_setup_cdb(hcb_p np, struct scsipi_xfer *xp, ccb_p cp)
{
	u32	cmd_ba;
	int	cmd_len;
	
	/*
	 *  CDB is 16 bytes max.
	 */
	if (xp->cmdlen > sizeof(cp->cdb_buf)) {
		sym_set_cam_status(cp->cam_ccb, CAM_REQ_INVALID);
		return -1;
	}

	cmd_len = xp->cmdlen;
	bcopy(xp->cmd, cp->cdb_buf, cmd_len);
	cmd_ba = CCB_BA (cp, cdb_buf[0]);

	cp->phys.cmd.addr	= cpu_to_scr(cmd_ba);
	cp->phys.cmd.size	= cpu_to_scr(cmd_len);

	return 0;
}

/*
 *  Setup everything for this IO.
 */
int sym_setup_data_and_start(hcb_p np, struct scsipi_xfer *xp, ccb_p cp)
{
	int dir, retv;
	tcb_p tp = &np->target[cp->target];
	lcb_p lp = sym_lp(np, tp, cp->lun);
	bus_dma_segment_t *psegs;
	int nsegs;

	/*
	 *  Set up the CDB.
	 */
	if (sym_setup_cdb(np, xp, cp))
		goto out_err;

	/*
	 *  Now deal with the data.
	 */
	cp->data_len = xp->datalen;

	/*
	 *  No direction means no data.
	 */
	dir = CAM_DIR_NONE;
	if (xs_ctl_data_in_isset(xp))
		dir = CAM_DIR_IN;
	else if (xs_ctl_data_out_isset(xp))
		dir = CAM_DIR_OUT;

	if (dir == CAM_DIR_NONE)
		goto no_data;

	if (!xs_ctl_data_uio_isset(xp)) {
		retv = bus_dmamap_load(np->s.data_dmat, cp->s.dmamap,
		                       xp->data, xp->datalen,
			               NULL, BUS_DMA_NOWAIT);
	}
	else {
		retv = bus_dmamap_load_uio(np->s.data_dmat, cp->s.dmamap,
			                   (struct uio *) xp->data,
		                           BUS_DMA_NOWAIT);
	}
	if (retv) {
		sym_set_cam_status(cp->cam_ccb, CAM_RESRC_UNAVAIL);
		goto out_err;
	}

	cp->s.dmamapped = (dir == CAM_DIR_IN) ? SYM_DMA_READ : SYM_DMA_WRITE;
	bus_dmamap_sync(np->s.data_dmat, cp->s.dmamap, 0,
			cp->s.dmamap->dm_mapsize,
			cp->s.dmamapped == SYM_DMA_READ ? 
				BUS_DMASYNC_PREREAD : BUS_DMASYNC_PREWRITE);

	/*
	 *  Build the data descriptor for the chip.
	 *  896 rev 1 requires to be careful about boundaries.
	 */
	psegs = cp->s.dmamap->dm_segs;
	nsegs = cp->s.dmamap->dm_nsegs;
	if (np->device_id == PCI_ID_SYM53C896 && np->revision_id <= 1)
		retv = sym_scatter_sg_physical(np, cp, psegs, nsegs);
	else
		retv = sym_fast_scatter_sg_physical(np, cp, psegs, nsegs);
	if (retv < 0) {
		sym_set_cam_status(cp->cam_ccb, CAM_REQ_TOO_BIG);
		goto out_err;
	}

	/*
	 *  Set data pointers.
	 */
no_data:
	sym_setup_data_pointers(np, cp,  dir);

	/*
	 *  When `#ifed 1', the code below makes the driver 
	 *  panic on the first attempt to write to a SCSI device.
	 *  It is the first test we want to do after a driver 
	 *  change that does not seem obviously safe. :)
	 */
#if 0
	switch (cp->cdb_buf[0]) {
	case 0x0A: case 0x2A: case 0xAA:
		panic("XXXXXXXXXXXXX WRITE NOT YET ALLOWED XXXXXXXXXXXXXX\n");
		MDELAY(10000);
		break;
	default:
		break;
	}
#endif
	/*
	 *  Activate this job.
	 */
	if (lp)
		sym_start_next_ccbs(np, lp, 2);
	else
		sym_put_start_queue(np, cp);

	return SUCCESSFULLY_QUEUED;
out_err:
	sym_free_ccb(np, cp);
	return COMPLETE;
}

/*
 *  Move the scatter list to our data block.
 */
static int 
sym_fast_scatter_sg_physical(hcb_p np, ccb_p cp, 
			     bus_dma_segment_t *psegs, int nsegs)
{
	struct sym_tblmove *data;
	bus_dma_segment_t *psegs2;

	if (nsegs > SYM_CONF_MAX_SG)
		return -1;

	data   = &cp->phys.data[SYM_CONF_MAX_SG-1];
	psegs2 = &psegs[nsegs-1];
	cp->segments = nsegs;

	while (1) {
		sym_build_sge(np, data, psegs2->ds_addr, psegs2->ds_len);
		if (DEBUG_FLAGS & DEBUG_SCATTER) {
			printf ("%s scatter: paddr=%lx len=%ld\n",
				sym_name(np), (long) psegs2->ds_addr,
				(long) psegs2->ds_len);
		}
		if (psegs2 != psegs) {
			--data;
			--psegs2;
			continue;
		}
		break;
	}
	return 0;
}

/*
 *  Scatter a SG list with physical addresses into bus addressable chunks.
 *  We need to ensure 16MB boundaries not to be crossed during DMA of 
 *  each segment, due to some chips being flawed.
 */
#define BOUND_MASK ((1UL<<24)-1)
static int
sym_scatter_sg_physical(hcb_p np, ccb_p cp, bus_dma_segment_t *psegs, int nsegs)
{
	bus_addr_t ps, pe, pn;
	u_long	k; 
	int s, t;

	s  = SYM_CONF_MAX_SG - 1;
	t  = nsegs - 1;
	ps = psegs[t].ds_addr;
	pe = ps + psegs[t].ds_len;

	while (s >= 0) {
		pn = (pe - 1) & ~BOUND_MASK;
		if (pn <= ps)
			pn = ps;
		k = pe - pn;
		if (DEBUG_FLAGS & DEBUG_SCATTER) {
			printf ("%s scatter: paddr=%lx len=%ld\n",
				sym_name(np), (long)pn, k);
		}
		sym_build_sge(np, &cp->phys.data[s], pn, k);
		--s;
		if (pn == ps) {
			if (--t < 0)
				break;
			ps = psegs[t].ds_addr;
			pe = ps + psegs[t].ds_len;
		}
		else
			pe = pn;
	}

	cp->segments = SYM_CONF_MAX_SG - 1 - s;

	return t >= 0 ? -1 : 0;
}
#undef BOUND_MASK

/*
 *  Shorten the IO to the maximum we can handle.
 */
static void sym_minphys(struct buf *bp)
{
        if (bp->b_bcount > ((SYM_CONF_MAX_SG - 1) * PAGE_SIZE)) {
                bp->b_bcount = ((SYM_CONF_MAX_SG - 1) * PAGE_SIZE);
        }
	minphys(bp);
}

/*
 *  Provide io control functions.
 */
static int      
sym_ioctl(struct scsipi_link *sc_link, u_long cmd, caddr_t addr,
	  int flag, struct proc *p)
{
	hcb_p np;
	int s, retv = ENOTTY;

	s = splbio();

	np = ((struct sym_data *)sc_link->adapter_softc)->np;

	switch (cmd) {
	case SCBUSIORESET:
		retv = 0;
		if (sym_reset_scsi_bus(np, 1))
			retv = EIO;
		break;
#if __NetBSD_Version__ >= 105000000
	case SCBUSACCEL:
	{
		struct scbusaccel_args *sp = (struct scbusaccel_args *)addr;
		tcb_p tp = &np->target[sp->sa_target];
		lcb_p lp;

		lp = sym_alloc_lcb(np, sp->sa_target, sp->sa_lun);
		if (lp) {
			lp->curr_flags = lp->user_flags;
			if (!(sp->sa_flags & SC_ACCEL_TAGS))
				lp->curr_flags &= ~SYM_TAGS_ENABLED;
			sym_setup_dev_queuing(np, tp, lp, sc_link);
		}

		tp->inq_byte7_valid |= (INQ7_WIDE16|INQ7_SYNC);

		if (!(sp->sa_flags & SC_ACCEL_WIDE))
			tp->inq_byte7_valid &= ~INQ7_WIDE16;

		if (!(sp->sa_flags & SC_ACCEL_SYNC))
			tp->inq_byte7_valid &= ~INQ7_SYNC;

		sym_update_trans_settings(np, tp);

		retv = 0;
		break;
	}
#endif
	default:
		break;
	}

	splx(s);
	return retv;
}

/*
 *  We provide the portable side with a 1 second clock.
 */
static void sym_timeout_handler(void *arg)
{
	int s = splbio();
#if __NetBSD_Version__ >= 105000000
	callout_reset(&((hcb_p) arg)->s.clock_ch, SYM_CLOCK_TICKS, 
		      sym_timeout_handler, arg);
#else
	timeout(sym_timeout_handler, arg, SYM_CLOCK_TICKS);
#endif
	sym_clock((hcb_p) arg);
	splx(s);
}

/*============= DRIVER INITIALISATION ==================*/

static int sym_pci_probe(struct device *, struct cfdata *, void *);
static void sym_pci_attach(struct device *, struct device *, void *);

struct cfattach sym_ca = {
	sizeof(struct sym_data), sym_pci_probe, sym_pci_attach
};

static struct scsipi_device sym_dev =
{
    NULL,                       /* Use default error handler */
    NULL,                       /* have a queue, served by this */
    NULL,                       /* have no async handler */
    NULL,                       /* Use default 'done' routine */
};

/*
 *  Look up the chip table.
 *
 *  Return a pointer to the chip entry if found, 
 *  zero otherwise.
 */
static struct sym_pci_chip *
sym_find_pci_chip(struct pci_attach_args *pa)
{
	if (PCI_VENDOR(pa->pa_id) != PCI_VENDOR_NCR)
		return 0;

	return sym_lookup_pci_chip_table(PCI_PRODUCT(pa->pa_id),
					 PCI_REVISION(pa->pa_class));
}

/*
 *  Tell upper layer if the chip is supported.
 */
static int
sym_pci_probe(struct device *parent, struct cfdata *match, void *aux)
{
	struct pci_attach_args *pa = aux;
	struct	sym_pci_chip *chip;

	chip = sym_find_pci_chip(pa);
	if (chip && sym_find_firmware(chip)) {
		return 1;
	}

	return 0;
}

/*
 *  Attach a sym53c8xx device.
 */
static int 
sym_pci_attach2(struct device *parent, struct device *self, void *aux)
{
	struct pci_attach_args *pa = aux;
	bus_addr_t bus_addr;
	pci_intr_handle_t ih;
	pcireg_t reg_type;
	struct	sym_pci_chip *chip;
	u32	pci_csr, pci_bhlcr;
	struct	sym_hcb *np = 0;
	struct	sym_fw *fw = 0;
	int 	i;

	/*
	 *  Only probed devices should be attached.
	 *  We just enjoy being paranoid. :)
	 */
	chip = sym_find_pci_chip(pa);
	if (chip == NULL || (fw = sym_find_firmware(chip)) == NULL)
		return (ENXIO);

	/*
	 *  Print the short name of the thing.
	 */
	printf(": <Symbios/LSI 53c%s>\n", chip->name);

	/*
	 *  Allocate immediately the host control block, 
	 *  since we are only expecting to succeed. :)
	 *  We keep track in the HCB of all the resources that 
	 *  are to be released on error.
	 */
	np = __sym_calloc_dma(pa->pa_dmat, sizeof(*np), "HCB");
	if (!np)
		goto attach_failed;

	/*
	 *  Move some useful data to our HBA structure.
	 */
	np->bus_dmat = pa->pa_dmat;
	np->s.data_dmat = pa->pa_dmat;
	np->s.device = self;
	((struct sym_data *) self)->np = np;
	np->s.pa_pc = pa->pa_pc;
	np->s.pa_tag = pa->pa_tag;

	/*
	 *  Copy some useful infos to the HCB.
	 */
	np->hcb_ba	 = vtobus(np);
	np->verbose	 = bootverbose;
	np->device_id	 = PCI_PRODUCT(pa->pa_device);
	np->revision_id  = PCI_REVISION(pa->pa_class);
	np->features	 = chip->features;
	np->clock_divn	 = chip->nr_divisor;
	np->maxoffs	 = chip->offset_max;
	np->maxburst	 = chip->burst_max;

	/*
	 *  Read and apply some fix-ups to the PCI COMMAND 
	 *  register. We want the chip to be enabled for:
	 *  - BUS mastering
	 *  - PCI parity checking (reporting would also be fine)
	 *  - Write And Invalidate.
	 */
	pci_csr = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_COMMAND_STATUS_REG);
	pci_csr |= PCI_COMMAND_MASTER_ENABLE;
	pci_csr |= PCI_COMMAND_INVALIDATE_ENABLE;
	pci_csr |= PCI_COMMAND_PARITY_ENABLE;
	pci_csr |= PCI_COMMAND_SERR_ENABLE;
	pci_conf_write(pa->pa_pc, pa->pa_tag, PCI_COMMAND_STATUS_REG, pci_csr);

	/*
	 *  Let the device know about the cache line size, 
	 *  if it doesn't yet.
	 */
	pci_bhlcr = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_BHLC_REG);
	if (!(PCI_CACHELINE(pci_bhlcr))) {
		printf("%s: CACHE LINE SIZE NOT CONFIGURED.\n", sym_name(np));
	}

	/*
	 *  Allocate/get/map/retrieve everything that deals with MMIO.
	 */
	reg_type = PCI_MAPREG_TYPE_MEM | PCI_MAPREG_MEM_TYPE_32BIT;
	if (chip->features & FE_64BIT)
		reg_type = PCI_MAPREG_TYPE_MEM | PCI_MAPREG_MEM_TYPE_64BIT;
	if (pci_mapreg_map(pa, SYM_PCI_MMIO, reg_type, 0,
			&np->s.mmio.tag, &np->s.mmio.bsh, &bus_addr, NULL)) {
		printf("%s: failed to map MMIO window\n", sym_name(np));
		goto attach_failed;
	}
	np->mmio_ba = (u32)bus_addr;

#ifdef	SYM_CONF_IOMAPPED
	/*
	 *  Allocate/get/map/retrieve everything that deals with IO.
	 */
	if (pci_mapreg_map(pa, SYM_PCI_IO, PCI_MAPREG_TYPE_IO, 0,
			&np->s.io.tag, &np->s.io.bsh, &bus_addr, NULL)) {
		printf("%s: failed to map IO window\n", sym_name(np));
		goto attach_failed;
	}
#endif

	/*
	 *  If the chip has RAM.
	 *  Allocate/get/map/retrieve the corresponding resources.
	 */
	if (np->features & (FE_RAM|FE_RAM8K)) {
		bus_size_t size;
		i = (chip->features & FE_64BIT)? SYM_PCI_RAM64 : SYM_PCI_RAM;
		if (pci_mapreg_map(pa, i, reg_type, 0,
			&np->s.ram.tag, &np->s.ram.bsh, &bus_addr, &size)) {
			printf("%s: failed to map RAM window\n", sym_name(np));
			goto attach_failed;
		}
		np->ram_ba = (u32)bus_addr;
	}

	/*
	 *  Deal with the interrupt stuff.
	 */
#if __NetBSD_Version__ >= 106000000
	if (pci_intr_map(pa, &ih))
#else
	if (pci_intr_map(pa->pa_pc, pa->pa_intrtag, pa->pa_intrpin,
	    pa->pa_intrline, &ih))
#endif
	{
		printf("%s: couldn't map interrupt\n", sym_name(np));
			goto attach_failed;
	}
	np->s.intr = pci_intr_establish(pa->pa_pc, ih, IPL_BIO, sym_intr, np);
	if (!np->s.intr) {
		printf("%s: couldn't establish interrupt\n", sym_name(np));
		goto attach_failed;
	}

	/*
	 *  Perform O/S independant stuff.
	 */
	if (sym_hcb_attach(np, fw))
		goto attach_failed;

	/*
	 *  Now deal with CAM or whatever looks like the SCSI 
	 *  access method on this platform.
	 *  Hopefully, we will succeed with that one.:)
	 */
	if (!sym_cam_attach(np))
		goto attach_failed;

	/*
	 *  Reset the SCSI BUS and start the SCRIPTS.
	 */
	(void)sym_reset_scsi_bus(np, 0);
	sym_start_up(np, 1);

	/*
	 *  Start the clock.
	 */
#if 1
#if __NetBSD_Version__ >= 105000000
	callout_init(&np->s.clock_ch);
#endif
	sym_timeout_handler((void *) np);
#endif

	/*
	 *  Sigh! we are done.
	 */
	config_found(np->s.device, &np->s.sc_link, scsiprint);
#if 0
	DEBUG_FREEZE("sym_pci_attach2: success - HIT THE RESET BUTTON!");
#endif
	return 0;

	/*
	 *  We have failed.
	 *  We will try to free all the resources we have 
	 *  allocated, but if we are a boot device, this 
	 *  will not help that much.;)
	 */
attach_failed:
	if (np)
		sym_pci_free(np);
	return ENXIO;
}

static void 
sym_pci_attach(struct device *parent, struct device *self, void *aux)
{
	(void) sym_pci_attach2(parent, self, aux);
}

/*
 *  Free everything that have been allocated for this device.
 */
static void sym_pci_free(hcb_p np)
{
	int s;

	s = splbio();

	/*
	 *  Cancel our clock (we are in full race here).
	 */
#if __NetBSD_Version__ >= 105000000
	callout_stop(&np->s.clock_ch);
#else
	untimeout(sym_timeout_handler, (void *) np);
#endif

	/*
	 *  First free CAM resources.
	 */
	sym_cam_free(np);

	/*
	 *  Free O/S independant resources.
	 */
	sym_hcb_free(np);

	/*
	 *  Now every should be quiet for us to 
	 *  free other resources, but I dunno how 
	 *  to do so for the moment. :)
	 */
	/* Free on-chip RAM related resources */
	/* Free MMIO related resources */
#ifdef	SYM_CONF_IOMAPPED
	/* Free IO related resources */
#endif
	/* Free IRQ/Interrupt related resources */

	/*
	 *  Free our control block.
	 */
	sym_mfree_dma(np, sizeof(*np), "HCB");

	splx(s);
}

/*
 *  Allocate CAM resources and register a bus to CAM.
 */
static int sym_cam_attach(hcb_p np)
{
	np->s.sc_adapter.scsipi_cmd     = sym_start_io;
	np->s.sc_adapter.scsipi_minphys = sym_minphys;
	np->s.sc_adapter.scsipi_ioctl   = sym_ioctl;

	np->s.sc_link.adapter_softc = np->s.device; /* At sym_data offset 0 */ 
	np->s.sc_link.scsipi_scsi.adapter_target = np->myaddr;
	np->s.sc_link.openings = 1;
	np->s.sc_link.scsipi_scsi.channel = SCSI_CHANNEL_ONLY_ONE;
	np->s.sc_link.scsipi_scsi.max_target = np->maxwide ? 15 : 7;
	np->s.sc_link.scsipi_scsi.max_lun = SYM_CONF_MAX_LUN-1;
	np->s.sc_link.type = BUS_SCSI;
	np->s.sc_link.adapter = &np->s.sc_adapter;

	np->s.sc_link.device = &sym_dev;
	np->s.sc_link.flags = 0;

	return 1;
}

/*
 *  Free everything that deals with CAM.
 *  Dunno how to do so for the moment.
 */
static void sym_cam_free(hcb_p np)
{
	return;
}
